import pytest
from copy import deepcopy
import json
from queue import Queue
from typing import Dict, List, cast

from vellum.utils.json_encoder import VellumJsonEncoder
from vellum.workflows.constants import undefined
from vellum.workflows.nodes.bases import BaseNode
from vellum.workflows.outputs.base import BaseOutputs
from vellum.workflows.state.base import BaseState
from vellum.workflows.state.delta import SetStateDelta, StateDelta
from vellum.workflows.types.code_execution_node_wrappers import DictWrapper


@pytest.fixture()
def mock_deepcopy(mocker):
    return mocker.patch("vellum.workflows.state.base.deepcopy")


@pytest.fixture
def mock_logger(mocker):
    return mocker.patch("vellum.workflows.state.base.logger")


class MockState(BaseState):
    foo: str
    nested_dict: Dict[str, int] = {}

    __snapshot_count__: int = 0

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.__snapshot_callback__ = lambda _, __: self.__mock_snapshot__()

    def __mock_snapshot__(self) -> None:
        self.__snapshot_count__ += 1

    def __deepcopy__(self, memo: dict) -> "MockState":
        new_state = cast(MockState, super().__deepcopy__(memo))
        new_state.__snapshot_count__ = 0
        return new_state


class MockNode(BaseNode):
    class ExternalInputs(BaseNode.ExternalInputs):
        message: str

    class Outputs(BaseOutputs):
        baz: str


MOCK_NODE_OUTPUT_ID = "a864331d-9b1c-43db-90e7-6b1998d8179d"


def test_state_snapshot__node_attribute_edit():
    # GIVEN an initial state instance
    state = MockState(foo="bar")
    assert state.__snapshot_count__ == 0

    # WHEN we edit an attribute
    state.foo = "baz"

    # THEN the snapshot is emitted
    assert state.__snapshot_count__ == 1


def test_state_snapshot__node_output_edit():
    # GIVEN an initial state instance
    state = MockState(foo="bar")
    assert state.__snapshot_count__ == 0

    # WHEN we add a Node Output to state
    for output in MockNode.Outputs:
        state.meta.node_outputs[output] = "hello"

    # THEN the snapshot is emitted
    assert state.__snapshot_count__ == 1


def test_state_snapshot__nested_dictionary_edit():
    # GIVEN an initial state instance
    state = MockState(foo="bar")
    assert state.__snapshot_count__ == 0

    # WHEN we edit a nested dictionary
    state.nested_dict["hello"] = 1

    # THEN the snapshot is emitted
    assert state.__snapshot_count__ == 1


def test_state_snapshot__external_input_edit():
    # GIVEN an initial state instance
    state = MockState(foo="bar")
    assert state.__snapshot_count__ == 0

    # WHEN we add an external input to state
    state.meta.external_inputs[MockNode.ExternalInputs.message] = "hello"

    # THEN the snapshot is emitted
    assert state.__snapshot_count__ == 1


def test_state_deepcopy():
    # GIVEN an initial state instance
    state = MockState(foo="bar")

    # AND we add a Node Output to state
    state.meta.node_outputs[MockNode.Outputs.baz] = "hello"

    # WHEN we deepcopy the state
    deepcopied_state = deepcopy(state)

    # THEN node outputs are deepcopied
    assert deepcopied_state.meta.node_outputs == state.meta.node_outputs


def test_state_deepcopy__with_node_output_updates():
    # GIVEN an initial state instance
    state = MockState(foo="bar")

    # AND we add a Node Output to state
    state.meta.node_outputs[MockNode.Outputs.baz] = "hello"

    # AND we deepcopy the state
    deepcopied_state = deepcopy(state)

    # AND we update the original state
    state.meta.node_outputs[MockNode.Outputs.baz] = "world"

    # THEN the copied state is not updated
    assert deepcopied_state.meta.node_outputs[MockNode.Outputs.baz] == "hello"

    # AND the original state has had the correct number of snapshots
    assert state.__snapshot_count__ == 2

    # AND the copied state has had the correct number of snapshots
    assert deepcopied_state.__snapshot_count__ == 0


def test_state_json_serialization__with_node_output_updates():
    # GIVEN an initial state instance
    state = MockState(foo="bar")

    # AND we add a Node Output to state
    state.meta.node_outputs[MockNode.Outputs.baz] = "hello"

    # WHEN we serialize the state
    json_state = json.loads(json.dumps(state, cls=VellumJsonEncoder))

    # THEN the state is serialized correctly
    assert json_state["meta"]["node_outputs"] == {MOCK_NODE_OUTPUT_ID: "hello"}


def test_state_deepcopy__with_external_input_updates():
    # GIVEN an initial state instance
    state = MockState(foo="bar")

    # AND we add an external input to state
    state.meta.external_inputs[MockNode.ExternalInputs.message] = "hello"

    # AND we deepcopy the state
    deepcopied_state = deepcopy(state)

    # AND we update the original state
    state.meta.external_inputs[MockNode.ExternalInputs.message] = "world"

    # THEN the copied state is not updated
    assert deepcopied_state.meta.external_inputs[MockNode.ExternalInputs.message] == "hello"

    # AND the original state has had the correct number of snapshots
    assert state.__snapshot_count__ == 2

    # AND the copied state has had the correct number of snapshots
    assert deepcopied_state.__snapshot_count__ == 0


def test_state_json_serialization__with_queue():
    # GIVEN an initial state instance
    state = MockState(foo="bar")

    # AND we add a Node Output queue to state
    queue: Queue[str] = Queue()
    queue.put("test1")
    queue.put("test2")

    state.meta.node_outputs[MockNode.Outputs.baz] = queue

    # WHEN we serialize the state
    json_state = json.loads(json.dumps(state, cls=VellumJsonEncoder))

    # THEN the state is serialized correctly with the queue turned into a list
    assert json_state["meta"]["node_outputs"] == {MOCK_NODE_OUTPUT_ID: ["test1", "test2"]}


def test_state_snapshot__deepcopy_fails__logs_error(mock_deepcopy, mock_logger):
    # GIVEN an initial state instance
    class MockState(BaseState):
        foo: str

    state = MockState(foo="bar")

    # AND we have a snapshot callback that we are tracking
    snapshot_count = 0

    def __snapshot_callback__(state: "BaseState", deltas: List[StateDelta]) -> None:
        nonlocal snapshot_count
        snapshot_count += 1

    state.__snapshot_callback__ = __snapshot_callback__

    # AND deepcopy fails the first time but succeeds the second time
    deepcopy_side_effect_count = 0

    def side_effect(*args, **kwargs):
        nonlocal deepcopy_side_effect_count
        deepcopy_side_effect_count += 1
        if deepcopy_side_effect_count == 1:
            raise Exception("Failed to deepcopy")
        return deepcopy(args[0])

    mock_deepcopy.side_effect = side_effect

    # WHEN we snapshot the state twice
    state.__snapshot__(SetStateDelta(name="foo", delta="bar"))
    state.__snapshot__(SetStateDelta(name="foo", delta="baz"))

    # THEN we were able to invoke the callback once
    assert snapshot_count == 1

    # AND alert sentry once
    assert mock_logger.exception.call_count == 1


def test_state_deepcopy_handles_undefined_values():
    # GIVEN a state with undefined values in node outputs
    state = MockState(foo="bar")
    state.meta.node_outputs[MockNode.Outputs.baz] = DictWrapper({"foo": undefined})

    # WHEN we deepcopy the state
    deepcopied_state = deepcopy(state)

    # THEN the undefined values are preserved
    assert deepcopied_state.meta.node_outputs[MockNode.Outputs.baz] == {"foo": undefined}



from typing import TYPE_CHECKING, Any, Generic, Optional, Tuple, Type, TypeVar, cast

from pydantic import GetCoreSchemaHandler
from pydantic_core import core_schema

from vellum.workflows.constants import undefined
from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.exceptions import NodeException

if TYPE_CHECKING:
    from vellum.workflows.inputs.base import BaseInputs
    from vellum.workflows.state.base import BaseState

_InputType = TypeVar("_InputType")


class ExternalInputReference(BaseDescriptor[_InputType], Generic[_InputType]):

    def __init__(
        self,
        *,
        name: str,
        types: Tuple[Type[_InputType], ...],
        instance: Optional[_InputType],
        inputs_class: Type["BaseInputs"],
    ) -> None:
        super().__init__(name=name, types=types, instance=instance)
        self._inputs_class = inputs_class

    @property
    def inputs_class(self) -> Type["BaseInputs"]:
        return self._inputs_class

    def resolve(self, state: "BaseState") -> _InputType:
        external_input = state.meta.external_inputs.get(self)
        if external_input is not undefined:
            return cast(_InputType, external_input)

        if state.meta.parent:
            return self.resolve(state.meta.parent)

        raise NodeException(f"Missing required Node Input: {self._name}", code=WorkflowErrorCode.INVALID_INPUTS)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            return False

        # Check equality of the name
        base_equal = super().__eq__(other)
        if not base_equal:
            return False

        return self._inputs_class == other._inputs_class

    def __hash__(self) -> int:
        return hash((self._name, self._inputs_class))

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Type[Any], handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        return core_schema.is_instance_schema(cls)



from .api_node import BaseAPINode
from .inline_prompt_node import BaseInlinePromptNode
from .prompt_deployment_node import BasePromptDeploymentNode
from .search_node import BaseSearchNode

__all__ = [
    "BaseAPINode",
    "BaseInlinePromptNode",
    "BasePromptDeploymentNode",
    "BaseSearchNode",
]



from datetime import datetime
from typing import Optional

from vellum.workflows.triggers.base import BaseTrigger


class ScheduleTrigger(BaseTrigger):
    """
    Trigger representing time-based workflow invocation.
    Supports Cron-based schedules (e.g., "0 9 * * MON" for every Monday at 9am)
    """

    current_run_at: datetime
    next_run_at: datetime

    class Config:
        cron: str
        timezone: Optional[str] = None



from functools import reduce
from uuid import UUID
from typing import TYPE_CHECKING, Any, List, Literal, Optional, Sequence, Type, Union

from pydantic import ConfigDict, ValidationError

from vellum.client.core.pydantic_utilities import UniversalBaseModel
from vellum.client.types.array_vellum_value import ArrayVellumValue
from vellum.client.types.vellum_value import VellumValue
from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.exceptions import WorkflowInitializationException
from vellum.workflows.outputs.base import BaseOutputs
from vellum.workflows.references.constant import ConstantValueReference

if TYPE_CHECKING:
    from vellum.workflows import BaseWorkflow

import logging

logger = logging.getLogger(__name__)


class _RawLogicalCondition(UniversalBaseModel):
    type: Literal["LOGICAL_CONDITION"] = "LOGICAL_CONDITION"
    lhs_variable_id: UUID
    operator: Literal["==", ">", ">=", "<", "<=", "!="]
    rhs_variable_id: UUID


class _RawLogicalConditionGroup(UniversalBaseModel):
    type: Literal["LOGICAL_CONDITION_GROUP"] = "LOGICAL_CONDITION_GROUP"
    conditions: List["_RawLogicalExpression"]
    combinator: Literal["AND", "OR"]
    negated: bool


_RawLogicalExpression = Union[_RawLogicalCondition, _RawLogicalConditionGroup]


class _RawLogicalExpressionVariable(UniversalBaseModel):
    id: UUID


class _RawMockWorkflowNodeExecutionConstantValuePointer(_RawLogicalExpressionVariable):
    type: Literal["CONSTANT_VALUE"] = "CONSTANT_VALUE"
    variable_value: VellumValue


class _RawMockWorkflowNodeExecutionNodeExecutionCounterPointer(_RawLogicalExpressionVariable):
    type: Literal["EXECUTION_COUNTER"] = "EXECUTION_COUNTER"
    node_id: UUID


class _RawMockWorkflowNodeExecutionInputVariablePointer(_RawLogicalExpressionVariable):
    type: Literal["INPUT_VARIABLE"] = "INPUT_VARIABLE"
    input_variable_id: UUID


class _RawMockWorkflowNodeExecutionNodeOutputPointer(_RawLogicalExpressionVariable):
    type: Literal["NODE_OUTPUT"] = "NODE_OUTPUT"
    node_id: UUID
    input_id: UUID


class _RawMockWorkflowNodeExecutionNodeInputPointer(_RawLogicalExpressionVariable):
    type: Literal["NODE_INPUT"] = "NODE_INPUT"
    node_id: UUID
    input_id: UUID


_RawMockWorkflowNodeExecutionValuePointer = Union[
    _RawMockWorkflowNodeExecutionConstantValuePointer,
    _RawMockWorkflowNodeExecutionNodeExecutionCounterPointer,
    _RawMockWorkflowNodeExecutionInputVariablePointer,
    _RawMockWorkflowNodeExecutionNodeOutputPointer,
    _RawMockWorkflowNodeExecutionNodeInputPointer,
]


class _RawMockWorkflowNodeWhenCondition(UniversalBaseModel):
    expression: _RawLogicalExpression
    variables: List[_RawMockWorkflowNodeExecutionValuePointer]


class _RawMockWorkflowNodeThenOutput(UniversalBaseModel):
    output_id: UUID
    value: _RawMockWorkflowNodeExecutionValuePointer


class _RawMockWorkflowNodeExecution(UniversalBaseModel):
    when_condition: _RawMockWorkflowNodeWhenCondition
    then_outputs: List[_RawMockWorkflowNodeThenOutput]


class _RawMockWorkflowNodeConfig(UniversalBaseModel):
    type: Literal["WORKFLOW_NODE_OUTPUT"] = "WORKFLOW_NODE_OUTPUT"
    node_id: UUID
    mock_executions: List[_RawMockWorkflowNodeExecution]


class MockNodeExecution(UniversalBaseModel):
    when_condition: BaseDescriptor
    then_outputs: BaseOutputs

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @staticmethod
    def validate_all(
        raw_mock_workflow_node_configs: Optional[List[Any]],
        workflow: Type["BaseWorkflow"],
    ) -> Optional[List["MockNodeExecution"]]:
        if not raw_mock_workflow_node_configs:
            return None

        ArrayVellumValue.model_rebuild()
        try:
            mock_workflow_node_configs = [
                _RawMockWorkflowNodeConfig.model_validate(raw_mock_workflow_node_config)
                for raw_mock_workflow_node_config in raw_mock_workflow_node_configs
            ]
        except ValidationError as e:
            raise WorkflowInitializationException(
                message="Failed to validate mock node executions",
                code=WorkflowErrorCode.INVALID_INPUTS,
                workflow_definition=workflow,
            ) from e

        nodes = {node.__id__: node for node in workflow.get_nodes()}
        node_output_name_by_id = {
            node.__output_ids__[output.name]: output.name for node in workflow.get_nodes() for output in node.Outputs
        }

        # We need to support the old way that the Vellum App's WorkflowRunner used to define Node Mocks in order to
        # avoid needing to update the mock resolution strategy that it and the frontend uses. The path towards
        # cleaning this up will go as follows:
        # 1. Release Mock support in SDK-Enabled Workflows
        # 2. Deprecate Mock support in non-SDK enabled Workflows, encouraging users to migrate to SDK-enabled Workflows
        # 3. Remove the old mock resolution strategy
        # 4. Update this SDK to handle the new mock resolution strategy with WorkflowValueDescriptors
        # 5. Cutover the Vellum App to the new mock resolution strategy
        # 6. Remove the old mock resolution strategy from this SDK
        def _translate_raw_logical_expression(
            raw_logical_expression: _RawLogicalExpression,
            raw_variables: List[_RawMockWorkflowNodeExecutionValuePointer],
        ) -> BaseDescriptor:
            if raw_logical_expression.type == "LOGICAL_CONDITION":
                return _translate_raw_logical_condition(raw_logical_expression, raw_variables)
            else:
                return _translate_raw_logical_condition_group(raw_logical_expression, raw_variables)

        def _translate_raw_logical_condition_group(
            raw_logical_condition_group: _RawLogicalConditionGroup,
            raw_variables: List[_RawMockWorkflowNodeExecutionValuePointer],
        ) -> BaseDescriptor:
            if not raw_logical_condition_group.conditions:
                return ConstantValueReference(True)

            conditions = [
                _translate_raw_logical_expression(condition, raw_variables)
                for condition in raw_logical_condition_group.conditions
            ]
            return reduce(
                lambda acc, condition: (
                    acc and condition if raw_logical_condition_group.combinator == "AND" else acc or condition
                ),
                conditions,
            )

        def _translate_raw_logical_condition(
            raw_logical_condition: _RawLogicalCondition,
            raw_variables: List[_RawMockWorkflowNodeExecutionValuePointer],
        ) -> BaseDescriptor:
            variable_by_id = {v.id: v for v in raw_variables}
            lhs = _translate_raw_logical_expression_variable(variable_by_id[raw_logical_condition.lhs_variable_id])
            rhs = _translate_raw_logical_expression_variable(variable_by_id[raw_logical_condition.rhs_variable_id])
            if raw_logical_condition.operator == ">":
                return lhs.greater_than(rhs)
            elif raw_logical_condition.operator == ">=":
                return lhs.greater_than_or_equal_to(rhs)
            elif raw_logical_condition.operator == "<":
                return lhs.less_than(rhs)
            elif raw_logical_condition.operator == "<=":
                return lhs.less_than_or_equal_to(rhs)
            elif raw_logical_condition.operator == "==":
                return lhs.equals(rhs)
            elif raw_logical_condition.operator == "!=":
                return lhs.does_not_equal(rhs)
            else:
                raise WorkflowInitializationException(
                    message=f"Unsupported logical operator: {raw_logical_condition.operator}",
                    workflow_definition=workflow,
                )

        def _translate_raw_logical_expression_variable(
            raw_variable: _RawMockWorkflowNodeExecutionValuePointer,
        ) -> BaseDescriptor:
            if raw_variable.type == "CONSTANT_VALUE":
                return ConstantValueReference(raw_variable.variable_value.value)
            elif raw_variable.type == "EXECUTION_COUNTER":
                node = nodes[raw_variable.node_id]
                return node.Execution.count
            else:
                raise WorkflowInitializationException(
                    message=f"Unsupported logical expression type: {raw_variable.type}",
                    workflow_definition=workflow,
                )

        mock_node_executions = []
        for mock_workflow_node_config in mock_workflow_node_configs:
            for mock_execution in mock_workflow_node_config.mock_executions:
                try:
                    when_condition = _translate_raw_logical_expression(
                        mock_execution.when_condition.expression,
                        mock_execution.when_condition.variables,
                    )

                    then_outputs = nodes[mock_workflow_node_config.node_id].Outputs()
                    for then_output in mock_execution.then_outputs:
                        node_output_name = node_output_name_by_id.get(then_output.output_id)
                        if node_output_name is None:
                            node_id = mock_workflow_node_config.node_id
                            raise WorkflowInitializationException(
                                message=f"Output {then_output.output_id} not found in node {node_id}",
                                workflow_definition=workflow,
                            )

                        resolved_output_reference = _translate_raw_logical_expression_variable(then_output.value)
                        if isinstance(resolved_output_reference, ConstantValueReference):
                            setattr(
                                then_outputs,
                                node_output_name,
                                resolved_output_reference._value,
                            )
                        else:
                            ref_type = type(resolved_output_reference)
                            raise WorkflowInitializationException(
                                message=f"Unsupported resolved output reference type: {ref_type}",
                                workflow_definition=workflow,
                            )

                    mock_node_executions.append(
                        MockNodeExecution(
                            when_condition=when_condition,
                            then_outputs=then_outputs,
                        )
                    )
                except Exception as e:
                    logger.exception("Failed to validate mock node execution", exc_info=e)
                    continue

        return mock_node_executions


MockNodeExecutionArg = Sequence[Union[BaseOutputs, MockNodeExecution]]



# This file was auto-generated by Fern from our API Definition.

import typing

from .prompt_request_audio_input import PromptRequestAudioInput
from .prompt_request_chat_history_input import PromptRequestChatHistoryInput
from .prompt_request_document_input import PromptRequestDocumentInput
from .prompt_request_image_input import PromptRequestImageInput
from .prompt_request_json_input import PromptRequestJsonInput
from .prompt_request_string_input import PromptRequestStringInput
from .prompt_request_video_input import PromptRequestVideoInput

PromptRequestInput = typing.Union[
    PromptRequestStringInput,
    PromptRequestJsonInput,
    PromptRequestChatHistoryInput,
    PromptRequestAudioInput,
    PromptRequestVideoInput,
    PromptRequestImageInput,
    PromptRequestDocumentInput,
]



# This file was auto-generated by Fern from our API Definition.

import typing

import httpx

from ..environment import VellumEnvironment
from ..types.api_version_enum import ApiVersionEnum
from .http_client import AsyncHttpClient, HttpClient


class BaseClientWrapper:
    def __init__(
        self,
        *,
        api_version: typing.Optional[ApiVersionEnum] = None,
        api_key: str,
        headers: typing.Optional[typing.Dict[str, str]] = None,
        environment: VellumEnvironment,
        timeout: typing.Optional[float] = None,
    ):
        self._api_version = api_version
        self.api_key = api_key
        self._headers = headers
        self._environment = environment
        self._timeout = timeout

    def get_headers(self) -> typing.Dict[str, str]:
        headers: typing.Dict[str, str] = {
            "User-Agent": "vellum-ai/1.9.9",
            "X-Fern-Language": "Python",
            "X-Fern-SDK-Name": "vellum-ai",
            "X-Fern-SDK-Version": "1.9.9",
            **(self.get_custom_headers() or {}),
        }
        if self._api_version is not None:
            headers["X-API-Version"] = self._api_version
        headers["X-API-KEY"] = self.api_key
        return headers

    def get_custom_headers(self) -> typing.Optional[typing.Dict[str, str]]:
        return self._headers

    def get_environment(self) -> VellumEnvironment:
        return self._environment

    def get_timeout(self) -> typing.Optional[float]:
        return self._timeout


class SyncClientWrapper(BaseClientWrapper):
    def __init__(
        self,
        *,
        api_version: typing.Optional[ApiVersionEnum] = None,
        api_key: str,
        headers: typing.Optional[typing.Dict[str, str]] = None,
        environment: VellumEnvironment,
        timeout: typing.Optional[float] = None,
        httpx_client: httpx.Client,
    ):
        super().__init__(
            api_version=api_version, api_key=api_key, headers=headers, environment=environment, timeout=timeout
        )
        self.httpx_client = HttpClient(
            httpx_client=httpx_client, base_headers=self.get_headers, base_timeout=self.get_timeout
        )


class AsyncClientWrapper(BaseClientWrapper):
    def __init__(
        self,
        *,
        api_version: typing.Optional[ApiVersionEnum] = None,
        api_key: str,
        headers: typing.Optional[typing.Dict[str, str]] = None,
        environment: VellumEnvironment,
        timeout: typing.Optional[float] = None,
        httpx_client: httpx.AsyncClient,
    ):
        super().__init__(
            api_version=api_version, api_key=api_key, headers=headers, environment=environment, timeout=timeout
        )
        self.httpx_client = AsyncHttpClient(
            httpx_client=httpx_client, base_headers=self.get_headers, base_timeout=self.get_timeout
        )



from typing import Union

from vellum.workflows.references.external_input import ExternalInputReference
from vellum.workflows.references.workflow_input import WorkflowInputReference

# Note: We may want to consolidate these two into a single descriptor.
InputReference = Union[ExternalInputReference, WorkflowInputReference]



from uuid import uuid4

from .inner_files.inner_class import InnerClass


class BaseClass:
    result = None

    def __init__(self):
        self.id = uuid4()
        self.inner_class = InnerClass

    def run(self):
        self.result = self.inner_class().introduce()
        return self.result



from deepdiff import DeepDiff

from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display

from tests.workflows.basic_tool_calling_node.workflow import BasicToolCallingNodeWorkflow


def test_serialize_workflow():
    # GIVEN a Workflow that uses a generic node
    # WHEN we serialize it
    workflow_display = get_workflow_display(workflow_class=BasicToolCallingNodeWorkflow)

    serialized_workflow: dict = workflow_display.serialize()
    # THEN we should get a serialized representation of the Workflow
    assert serialized_workflow.keys() == {
        "workflow_raw_data",
        "input_variables",
        "state_variables",
        "output_variables",
    }

    # AND its input variables should be what we expect
    input_variables = serialized_workflow["input_variables"]
    assert len(input_variables) == 1

    # AND its output variables should be what we expect
    output_variables = serialized_workflow["output_variables"]
    assert len(output_variables) == 2
    assert not DeepDiff(
        [
            {"id": "8e7c0147-930d-4b7f-b6b1-6d79641cd3eb", "key": "text", "type": "STRING"},
            {"id": "01a07e6d-7269-4f45-8b44-ef0227a2e88d", "key": "chat_history", "type": "CHAT_HISTORY"},
        ],
        output_variables,
        ignore_order=True,
    )

    # AND its raw data should be what we expect
    workflow_raw_data = serialized_workflow["workflow_raw_data"]
    tool_calling_node = workflow_raw_data["nodes"][1]
    assert tool_calling_node == {
        "id": "649a81cf-ce93-47c1-aa0f-e7a58a0cba8c",
        "label": "Get Current Weather Node",
        "type": "GENERIC",
        "should_file_merge": True,
        "display_data": {
            "position": {"x": 200.0, "y": -50.0},
            "comment": {
                "expanded": True,
                "value": "\n    A tool calling node that calls the get_current_weather function.\n    ",
            },
        },
        "base": {
            "name": "ToolCallingNode",
            "module": ["vellum", "workflows", "nodes", "displayable", "tool_calling_node", "node"],
        },
        "definition": {
            "name": "GetCurrentWeatherNode",
            "module": ["tests", "workflows", "basic_tool_calling_node", "workflow"],
        },
        "trigger": {"id": "fa785565-2906-433c-914a-79c574bcced4", "merge_behavior": "AWAIT_ATTRIBUTES"},
        "ports": [{"id": "8ae76650-6658-441e-aab2-dbdc62dc6a48", "name": "default", "type": "DEFAULT"}],
        "adornments": None,
        "attributes": [
            {
                "id": "6991a1bf-6a1d-4b0a-8e67-39505b324216",
                "name": "ml_model",
                "value": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "gpt-4o-mini"}},
            },
            {
                "id": "74ce80ae-4f12-48fd-a9b7-412532e383ec",
                "name": "blocks",
                "value": {
                    "type": "CONSTANT_VALUE",
                    "value": {
                        "type": "JSON",
                        "value": [
                            {
                                "block_type": "CHAT_MESSAGE",
                                "state": None,
                                "cache_config": None,
                                "chat_role": "SYSTEM",
                                "chat_source": None,
                                "chat_message_unterminated": None,
                                "blocks": [
                                    {
                                        "block_type": "RICH_TEXT",
                                        "state": None,
                                        "cache_config": None,
                                        "blocks": [
                                            {
                                                "block_type": "PLAIN_TEXT",
                                                "state": None,
                                                "cache_config": None,
                                                "text": "You are a weather expert",
                                            }
                                        ],
                                    }
                                ],
                            },
                            {
                                "block_type": "CHAT_MESSAGE",
                                "state": None,
                                "cache_config": None,
                                "chat_role": "USER",
                                "chat_source": None,
                                "chat_message_unterminated": None,
                                "blocks": [
                                    {
                                        "block_type": "RICH_TEXT",
                                        "state": None,
                                        "cache_config": None,
                                        "blocks": [
                                            {
                                                "block_type": "VARIABLE",
                                                "state": None,
                                                "cache_config": None,
                                                "input_variable": "aa87c0db-68c0-4c9b-9ad0-ddb48bb169eb",
                                            }
                                        ],
                                    }
                                ],
                            },
                        ],
                    },
                },
            },
            {
                "id": "1e3cd0b3-5657-42bd-8a71-48bdf9e9b835",
                "name": "functions",
                "value": {
                    "type": "CONSTANT_VALUE",
                    "value": {
                        "type": "JSON",
                        "value": [
                            {
                                "type": "CODE_EXECUTION",
                                "name": "get_current_weather",
                                "description": "\n    Get the current weather in a given location.\n    ",
                                "definition": {
                                    "state": None,
                                    "cache_config": None,
                                    "name": "get_current_weather",
                                    "description": "\n    Get the current weather in a given location.\n    ",
                                    "parameters": {
                                        "type": "object",
                                        "properties": {
                                            "location": {
                                                "type": "string",
                                                "description": "The location to get the weather for",
                                            },
                                            "unit": {"type": "string", "description": "The unit of temperature"},
                                        },
                                        "required": ["location", "unit"],
                                    },
                                    "inputs": None,
                                    "forced": None,
                                    "strict": None,
                                },
                                "src": 'import math\nfrom typing import Annotated\n\n\ndef get_current_weather(\n    location: Annotated[str, "The location to get the weather for"], unit: Annotated[str, "The unit of temperature"]\n) -> str:\n    """\n    Get the current weather in a given location.\n    """\n    return f"The current weather in {location} is sunny with a temperature of {get_temperature(70.1)} degrees {unit}."\n\n\ndef get_temperature(temperature: float) -> int:\n    """\n    Get the temperature in a given location.\n    """\n    return math.floor(temperature)\n',  # noqa: E501
                            }
                        ],
                    },
                },
            },
            {
                "id": "e554e8c1-a270-4fbf-951f-fd4aca4afe9b",
                "name": "prompt_inputs",
                "value": {
                    "type": "DICTIONARY_REFERENCE",
                    "entries": [
                        {
                            "id": "aa87c0db-68c0-4c9b-9ad0-ddb48bb169eb",
                            "key": "question",
                            "value": {
                                "type": "WORKFLOW_INPUT",
                                "input_variable_id": "a70e0597-0076-4a17-a8d2-c8d6b4a05bac",
                            },
                        }
                    ],
                },
            },
            {
                "id": "ef504b60-5b94-43d0-b31d-534309a52ff2",
                "name": "parameters",
                "value": {
                    "type": "CONSTANT_VALUE",
                    "value": {
                        "type": "JSON",
                        "value": {
                            "stop": [],
                            "temperature": 0.0,
                            "max_tokens": 4096.0,
                            "top_p": 1.0,
                            "top_k": 0.0,
                            "frequency_penalty": 0.0,
                            "presence_penalty": 0.0,
                            "logit_bias": None,
                            "custom_parameters": {"mode": "initial"},
                        },
                    },
                },
            },
            {
                "id": "83546f2d-f531-4773-b87b-aeb104b9218d",
                "name": "max_prompt_iterations",
                "value": {"type": "CONSTANT_VALUE", "value": {"type": "NUMBER", "value": 25.0}},
            },
            {
                "id": "bb8b8427-db2c-4115-b26b-c30dc705787a",
                "name": "settings",
                "value": {"type": "CONSTANT_VALUE", "value": {"type": "JSON", "value": None}},
            },
        ],
        "outputs": [
            {"id": "d0d0cd85-8546-4d90-bdc1-86c751ba04e5", "name": "text", "type": "STRING", "value": None},
            {
                "id": "33a737c2-d347-48d4-bf62-e36c06aebf0c",
                "name": "chat_history",
                "type": "CHAT_HISTORY",
                "value": None,
            },
        ],
    }



from abc import ABC, abstractmethod
import importlib.abc
from io import StringIO
from typing import Optional


class BaseWorkflowFinder(importlib.abc.MetaPathFinder, ABC):
    """
    Abstract base class for workflow finders that support custom error message formatting.
    """

    @abstractmethod
    def format_error_message(self, error_message: str) -> str:
        """
        Format an error message to be more user-friendly.

        Args:
            error_message: The original error message

        Returns:
            The formatted error message
        """
        pass

    @abstractmethod
    def virtual_open(self, file_path: str) -> Optional[StringIO]:
        """
        Open a virtual file if it exists in this finder's namespace.

        Args:
            file_path: The absolute file path to open

        Returns:
            A StringIO object containing the file contents, or None if the file is not found
        """
        pass



from typing import Any, ClassVar, Dict, Generic, Iterator, List, Optional, Set, Union

from vellum import ChatMessage, PromptBlock, PromptOutput
from vellum.client.types.prompt_parameters import PromptParameters
from vellum.client.types.prompt_settings import PromptSettings
from vellum.client.types.string_chat_message_content import StringChatMessageContent
from vellum.prompts.constants import DEFAULT_PROMPT_PARAMETERS
from vellum.workflows.context import execution_context, get_parent_context
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.events.workflow import is_workflow_event
from vellum.workflows.exceptions import NodeException
from vellum.workflows.graph.graph import Graph
from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.nodes.bases import BaseNode
from vellum.workflows.nodes.displayable.tool_calling_node.state import ToolCallingState
from vellum.workflows.nodes.displayable.tool_calling_node.utils import (
    create_else_node,
    create_function_node,
    create_mcp_tool_node,
    create_router_node,
    create_tool_prompt_node,
    get_function_name,
)
from vellum.workflows.outputs.base import BaseOutput, BaseOutputs
from vellum.workflows.state.context import WorkflowContext
from vellum.workflows.types.core import EntityInputsInterface
from vellum.workflows.types.definition import MCPServer, Tool
from vellum.workflows.types.generics import StateType
from vellum.workflows.utils.functions import compile_mcp_tool_definition, get_mcp_tool_name
from vellum.workflows.workflows.event_filters import all_workflow_event_filter


class ToolCallingNode(BaseNode[StateType], Generic[StateType]):
    """
    A Node that dynamically invokes the provided functions to the underlying Prompt

    Attributes:
        ml_model: str - The model to use for tool calling (e.g., "gpt-4o-mini")
        blocks: List[PromptBlock] - The prompt blocks to use (same format as InlinePromptNode)
        functions: List[Tool] - The functions that can be called
        prompt_inputs: Optional[EntityInputsInterface] - Mapping of input variable names to values
        parameters: PromptParameters - The parameters for the Prompt
        max_prompt_iterations: Optional[int] - Maximum number of prompt iterations before stopping
    """

    ml_model: ClassVar[str] = "gpt-4o-mini"
    blocks: ClassVar[List[Union[PromptBlock, Dict[str, Any]]]] = []
    functions: ClassVar[List[Tool]] = []
    prompt_inputs: ClassVar[Optional[EntityInputsInterface]] = None
    parameters: PromptParameters = DEFAULT_PROMPT_PARAMETERS
    max_prompt_iterations: ClassVar[Optional[int]] = 25
    settings: ClassVar[Optional[Union[PromptSettings, Dict[str, Any]]]] = None

    class Outputs(BaseOutputs):
        """
        The outputs of the ToolCallingNode.

        text: The final text response after tool calling
        chat_history: The complete chat history including tool calls
        """

        text: str
        chat_history: List[ChatMessage]

    def run(self) -> Iterator[BaseOutput]:
        """
        Run the tool calling workflow with streaming support.

        This dynamically builds a graph with router and function nodes,
        then executes the workflow and streams chat_history updates.
        """

        self._build_graph()

        with execution_context(parent_context=get_parent_context()):

            from vellum.workflows.workflows.base import BaseWorkflow

            class ToolCallingWorkflow(BaseWorkflow[BaseInputs, ToolCallingState]):
                graph = self._graph
                is_dynamic = True

                class Outputs(BaseWorkflow.Outputs):
                    text: str = self.tool_prompt_node.Outputs.text
                    chat_history: List[ChatMessage] = ToolCallingState.chat_history
                    results: List[PromptOutput] = self.tool_prompt_node.Outputs.results

            subworkflow = ToolCallingWorkflow(
                parent_state=self.state,
                context=WorkflowContext.create_from(self._context),
            )

            subworkflow_stream = subworkflow.stream(
                event_filter=all_workflow_event_filter,
                node_output_mocks=self._context._get_all_node_output_mocks(),
            )

        outputs: Optional[BaseOutputs] = None
        exception: Optional[NodeException] = None
        fulfilled_output_names: Set[str] = set()

        for event in subworkflow_stream:
            self._context._emit_subworkflow_event(event)

            if not is_workflow_event(event):
                continue
            if event.workflow_definition != ToolCallingWorkflow:
                continue

            if event.name == "workflow.execution.streaming":
                if event.output.name == "results":
                    if event.output.is_streaming:
                        if isinstance(event.output.delta, str):
                            text_message = ChatMessage(
                                text=event.output.delta,
                                role="ASSISTANT",
                                content=StringChatMessageContent(type="STRING", value=event.output.delta),
                                source=None,
                            )
                            yield BaseOutput(
                                name="chat_history",
                                delta=[text_message],
                            )
                elif event.output.name == "chat_history":
                    if event.output.is_fulfilled:
                        fulfilled_output_names.add(event.output.name)
                    yield event.output
                elif event.output.name == "text":
                    if event.output.is_fulfilled:
                        fulfilled_output_names.add(event.output.name)
                    yield event.output
            elif event.name == "workflow.execution.fulfilled":
                outputs = event.outputs
            elif event.name == "workflow.execution.rejected":
                exception = NodeException.of(event.error)

        if exception:
            # In the case of the Prompt Node receiving invalid inputs or schemas, that is a bug with the
            # internals of the Agent Node. In the future, we should distinguish further by also checking
            # that this exception came from the Tool Calling Node's prompt node instead of anything
            # internal to a user defined tool.
            if exception.code == WorkflowErrorCode.INVALID_INPUTS:
                raise NodeException(
                    message="Internal server error",
                    code=WorkflowErrorCode.INTERNAL_ERROR,
                    raw_data=exception.raw_data,
                    stacktrace=exception.stacktrace,
                ) from exception
            raise exception

        if outputs is None:
            raise NodeException(
                message="Expected to receive outputs from Tool Calling Workflow",
                code=WorkflowErrorCode.INVALID_OUTPUTS,
            )

        for output_descriptor, output_value in outputs:
            if output_descriptor.name not in fulfilled_output_names:
                yield BaseOutput(
                    name=output_descriptor.name,
                    value=output_value,
                )

    def _build_graph(self) -> None:
        # Get the process_parameters method if it exists on this class
        process_parameters_method = getattr(self.__class__, "process_parameters", None)
        process_blocks_method = getattr(self.__class__, "process_blocks", None)

        self.tool_prompt_node = create_tool_prompt_node(
            ml_model=self.ml_model,
            blocks=self.blocks,
            functions=self.functions,
            prompt_inputs=self.prompt_inputs,
            parameters=self.parameters,
            max_prompt_iterations=self.max_prompt_iterations,
            process_parameters_method=process_parameters_method,
            process_blocks_method=process_blocks_method,
            settings=self.settings,
        )

        # Create the router node (handles routing logic only)
        self.router_node = create_router_node(
            functions=self.functions,
            tool_prompt_node=self.tool_prompt_node,
        )

        self._function_nodes = {}
        for function in self.functions:
            if isinstance(function, MCPServer):
                tool_definitions = compile_mcp_tool_definition(function)
                for tool_definition in tool_definitions:
                    function_name = get_mcp_tool_name(tool_definition)

                    self._function_nodes[function_name] = create_mcp_tool_node(
                        tool_def=tool_definition,
                        tool_prompt_node=self.tool_prompt_node,
                    )
            else:
                function_name = get_function_name(function)

                self._function_nodes[function_name] = create_function_node(
                    function=function,
                    tool_prompt_node=self.tool_prompt_node,
                )

        graph: Graph = self.tool_prompt_node >> self.router_node

        for function_name, FunctionNodeClass in self._function_nodes.items():
            router_port = getattr(self.router_node.Ports, function_name)
            function_subgraph = router_port >> FunctionNodeClass >> self.router_node
            graph._extend_edges(function_subgraph.edges)

        else_node = create_else_node(self.tool_prompt_node)
        default_port_graph = self.router_node.Ports.default >> {
            else_node.Ports.loop_to_router >> self.router_node,  # More outputs to process
            else_node.Ports.loop_to_prompt >> self.tool_prompt_node,  # Need new prompt iteration
            else_node.Ports.end,  # Finished
        }
        graph._extend_edges(default_port_graph.edges)

        self._graph = graph



import pytest
from unittest.mock import MagicMock

from vellum.client.types.composio_execute_tool_response import ComposioExecuteToolResponse
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.exceptions import NodeException
from vellum.workflows.inputs import BaseInputs
from vellum.workflows.state import BaseState
from vellum.workflows.state.base import StateMeta

from ..node import WebSearchNode


@pytest.fixture
def base_node_setup(vellum_client):
    """Basic node setup with required inputs."""

    class Inputs(BaseInputs):
        query: str

    class State(BaseState):
        pass

    class TestableWebSearchNode(WebSearchNode):
        query = Inputs.query

    state = State(meta=StateMeta(workflow_inputs=Inputs(query="test query")))
    context = MagicMock()
    context.vellum_client = vellum_client
    node = TestableWebSearchNode(state=state, context=context)
    return node


def test_successful_search_with_results(base_node_setup):
    """Test successful SerpAPI search with typical organic results."""
    # GIVEN a mock Composio SerpAPI response with organic results
    mock_response_data = {
        "organic_results": [
            {
                "title": "First Result",
                "snippet": "This is the first search result snippet",
                "link": "https://example1.com",
                "position": 1,
            },
            {
                "title": "Second Result",
                "snippet": "This is the second search result snippet",
                "link": "https://example2.com",
                "position": 2,
            },
            {
                "title": "Third Result",
                "snippet": "This is the third search result snippet",
                "link": "https://example3.com",
                "position": 3,
            },
        ]
    }

    mock_response = ComposioExecuteToolResponse(provider="COMPOSIO", data=mock_response_data)
    base_node_setup._context.vellum_client.integrations.execute_integration_tool = MagicMock(return_value=mock_response)

    # WHEN we run the node
    outputs = base_node_setup.run()

    # THEN the text output should be properly formatted
    expected_text = (
        "First Result: This is the first search result snippet\n\n"
        "Second Result: This is the second search result snippet\n\n"
        "Third Result: This is the third search result snippet"
    )
    assert outputs.text == expected_text

    # AND URLs should be extracted correctly
    assert outputs.urls == ["https://example1.com", "https://example2.com", "https://example3.com"]

    # AND raw results should be preserved
    assert outputs.results == mock_response_data["organic_results"]

    # AND the Vellum client should have been called with the correct parameters
    base_node_setup._context.vellum_client.integrations.execute_integration_tool.assert_called_once_with(
        integration_name="SERPAPI",
        integration_provider="COMPOSIO",
        tool_name="SERPAPI_SEARCH",
        arguments={"query": "test query"},
    )


def test_authentication_error_401(base_node_setup):
    """Test authentication error raises NodeException with PROVIDER_ERROR."""
    base_node_setup._context.vellum_client.integrations.execute_integration_tool = MagicMock(
        side_effect=Exception("Authentication failed")
    )

    # WHEN we run the node
    with pytest.raises(NodeException) as exc_info:
        base_node_setup.run()

    # THEN it should raise the appropriate error
    assert exc_info.value.code == WorkflowErrorCode.PROVIDER_ERROR
    assert "Failed to execute web search" in str(exc_info.value)


def test_serpapi_error_in_response(base_node_setup):
    """Test SerpAPI error field in response raises NodeException."""
    # GIVEN SerpAPI returns an error in the response data
    mock_response = ComposioExecuteToolResponse(provider="COMPOSIO", data={"error": "Invalid search parameters"})
    base_node_setup._context.vellum_client.integrations.execute_integration_tool = MagicMock(return_value=mock_response)

    # WHEN we run the node
    with pytest.raises(NodeException) as exc_info:
        base_node_setup.run()

    # THEN it should raise the appropriate error
    assert exc_info.value.code == WorkflowErrorCode.PROVIDER_ERROR
    assert "Invalid search parameters" in str(exc_info.value)


def test_empty_query_validation(vellum_client):
    """Test empty query raises validation error."""

    # GIVEN a node with an empty query
    class TestNode(WebSearchNode):
        query = ""

    context = MagicMock()
    context.vellum_client = vellum_client
    node = TestNode(state=BaseState(meta=StateMeta(workflow_inputs=BaseInputs())), context=context)

    # WHEN we run the node
    with pytest.raises(NodeException) as exc_info:
        node.run()

    # THEN it should raise a validation error
    assert exc_info.value.code == WorkflowErrorCode.INVALID_INPUTS
    assert "Query is required" in str(exc_info.value)


def test_empty_organic_results(base_node_setup):
    """Test handling of empty search results."""
    # GIVEN Composio SerpAPI returns no organic results
    mock_response = ComposioExecuteToolResponse(provider="COMPOSIO", data={"organic_results": []})
    base_node_setup._context.vellum_client.integrations.execute_integration_tool = MagicMock(return_value=mock_response)

    # WHEN we run the node
    outputs = base_node_setup.run()

    # THEN all outputs should be empty
    assert outputs.text == ""
    assert outputs.urls == []
    assert outputs.results == []


def test_missing_fields_in_results(base_node_setup):
    """Test handling of missing title, snippet, or link fields."""
    # GIVEN Composio SerpAPI returns results with missing fields
    mock_response_data = {
        "organic_results": [
            {
                "title": "Only Title",
            },
            {"snippet": "Only snippet, no title or link"},
            {
                "title": "Title with link",
                "link": "https://example.com",
            },
            {"position": 4},
        ]
    }

    mock_response = ComposioExecuteToolResponse(provider="COMPOSIO", data=mock_response_data)
    base_node_setup._context.vellum_client.integrations.execute_integration_tool = MagicMock(return_value=mock_response)

    # WHEN we run the node
    outputs = base_node_setup.run()

    # THEN text should handle missing fields gracefully
    expected_text = "Only Title\n\n" "Only snippet, no title or link\n\n" "Title with link"
    assert outputs.text == expected_text

    # AND URLs should only include valid links
    assert outputs.urls == ["https://example.com"]



# This file was auto-generated by Fern from our API Definition.

import typing

IndexingStateEnum = typing.Union[
    typing.Literal["AWAITING_PROCESSING", "QUEUED", "INDEXING", "INDEXED", "FAILED"], typing.Any
]



import pytest
import json
from uuid import uuid4
from typing import Any, Iterator, List, cast

from vellum import ChatMessage
from vellum.client.core.api_error import ApiError
from vellum.client.types.execute_prompt_event import ExecutePromptEvent
from vellum.client.types.fulfilled_execute_prompt_event import FulfilledExecutePromptEvent
from vellum.client.types.function_call import FunctionCall
from vellum.client.types.function_call_vellum_value import FunctionCallVellumValue
from vellum.client.types.initiated_execute_prompt_event import InitiatedExecutePromptEvent
from vellum.client.types.prompt_output import PromptOutput
from vellum.client.types.string_chat_message_content import StringChatMessageContent
from vellum.client.types.string_vellum_value import StringVellumValue
from vellum.client.types.variable_prompt_block import VariablePromptBlock
from vellum.prompts.constants import DEFAULT_PROMPT_PARAMETERS
from vellum.workflows import BaseWorkflow
from vellum.workflows.constants import VellumIntegrationProviderType
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.exceptions import NodeException
from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.nodes.bases import BaseNode
from vellum.workflows.nodes.displayable.tool_calling_node.node import ToolCallingNode
from vellum.workflows.nodes.displayable.tool_calling_node.state import ToolCallingState
from vellum.workflows.nodes.displayable.tool_calling_node.utils import (
    create_function_node,
    create_router_node,
    create_tool_prompt_node,
)
from vellum.workflows.outputs.base import BaseOutputs
from vellum.workflows.ports.utils import validate_ports
from vellum.workflows.state.base import BaseState, StateMeta
from vellum.workflows.state.context import WorkflowContext
from vellum.workflows.types.definition import DeploymentDefinition, VellumIntegrationToolDefinition
from vellum.workflows.workflows.event_filters import all_workflow_event_filter


def first_function() -> str:
    return "first_function"


def second_function() -> str:
    return "second_function"


def test_router_node_port_ordering_with_multiple_tools():
    """
    Test that router node ports are created in the correct order: on_if, on_elif, ..., on_else.

    This test validates the fix for the bug where multiple tools would create multiple on_if
    ports instead of on_if followed by on_elif ports, which violates port validation rules.
    """

    # GIVEN three functions to ensure we test multiple elif cases
    def third_function() -> str:
        return "third_function"

    # AND a tool prompt node
    tool_prompt_node = create_tool_prompt_node(
        ml_model="test-model",
        blocks=[],
        functions=[first_function, second_function, third_function],
        prompt_inputs=None,
        parameters=DEFAULT_PROMPT_PARAMETERS,
    )

    # WHEN a router node is created with multiple functions
    router_node = create_router_node(
        functions=[first_function, second_function, third_function],
        tool_prompt_node=tool_prompt_node,
    )

    # THEN the first function port should be an on_if port
    first_function_port = getattr(router_node.Ports, "first_function")
    assert first_function_port._condition_type.value == "IF"

    # AND the second function port should be an on_elif port
    second_function_port = getattr(router_node.Ports, "second_function")
    assert second_function_port._condition_type.value == "ELIF"

    # AND the third function port should also be an on_elif port
    third_function_port = getattr(router_node.Ports, "third_function")
    assert third_function_port._condition_type.value == "ELIF"

    # AND the default port should be an on_else port
    default_port = getattr(router_node.Ports, "default")
    assert default_port._condition_type.value == "ELSE"

    # AND the ports should pass validation
    ports = [first_function_port, second_function_port, third_function_port, default_port]
    # This should not raise an exception
    validate_ports(ports)


def test_port_condition_match_function_name():
    """
    Test that the port condition correctly matches the function name.
    """
    # GIVEN a tool prompt node
    tool_prompt_node = create_tool_prompt_node(
        ml_model="test-model",
        blocks=[],
        functions=[first_function, second_function],
        prompt_inputs=None,
        parameters=DEFAULT_PROMPT_PARAMETERS,
    )

    # AND a router node that references the tool prompt node
    router_node = create_router_node(
        functions=[first_function, second_function],
        tool_prompt_node=tool_prompt_node,
    )

    # AND a state with a function call to the first function
    state = ToolCallingState(
        meta=StateMeta(
            node_outputs={
                tool_prompt_node.Outputs.results: [
                    FunctionCallVellumValue(
                        value=FunctionCall(
                            arguments={}, id="call_zp7pBQjGAOBCr7lo0AbR1HXT", name="first_function", state="FULFILLED"
                        ),
                    )
                ],
            },
        )
    )

    # WHEN the port condition is resolved
    # THEN the first function port should be true
    first_function_port = getattr(router_node.Ports, "first_function")
    assert first_function_port.resolve_condition(state) is True

    # AND the second function port should be false
    second_function_port = getattr(router_node.Ports, "second_function")
    assert second_function_port.resolve_condition(state) is False

    # AND the default port should be false
    default_port = getattr(router_node.Ports, "default")
    assert default_port.resolve_condition(state) is False


def test_tool_calling_node_inline_workflow_context():
    """
    Test that the tool calling node correctly passes the context to the inline workflow.
    This specifically tests that inline workflows receive the correct context.
    """

    # GIVEN a test workflow that captures its context
    class MyNode(BaseNode):
        class Outputs(BaseOutputs):
            generated_files: Any

        def run(self) -> Outputs:
            return self.Outputs(generated_files=self._context.generated_files)

    class MyWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = MyNode

        class Outputs(BaseOutputs):
            generated_files = MyNode.Outputs.generated_files

    # GIVEN a tool prompt node
    tool_prompt_node = create_tool_prompt_node(
        ml_model="test-model",
        blocks=[],
        functions=[MyWorkflow],
        prompt_inputs=None,
        parameters=DEFAULT_PROMPT_PARAMETERS,
    )

    # WHEN we create a function node for the workflow
    function_node_class = create_function_node(
        function=MyWorkflow,
        tool_prompt_node=tool_prompt_node,
    )

    # AND we create an instance with a context containing generated_files
    function_node = function_node_class()

    # Create a parent context with test data
    parent_context = WorkflowContext(
        generated_files={"script.py": "print('hello world')"},
    )
    function_node._context = parent_context

    # WHEN the function node runs
    outputs = list(function_node.run())

    # THEN the workflow should have run successfully
    assert outputs is not None

    # AND the chat history should contain a function response
    assert len(function_node.state.chat_history) == 1
    function_response = function_node.state.chat_history[0]
    assert function_response.role == "FUNCTION"

    # AND the response should contain the generated files
    assert isinstance(function_response.content, StringChatMessageContent)
    data = json.loads(function_response.content.value)
    assert data["generated_files"] == {"script.py": "print('hello world')"}


def test_deployment_definition_release_tag_defaults_to_latest():
    """
    Test that when creating a DeploymentDefinition without specifying release_tag,
    it defaults to "LATEST".
    """
    # WHEN we create a deployment definition without specifying release_tag
    deployment_config = DeploymentDefinition(deployment="test-deployment")

    # THEN the release_tag should default to "LATEST"
    assert deployment_config.release_tag == "LATEST"


def test_tool_calling_node_with_user_provided_chat_history_block(vellum_adhoc_prompt_client):
    """
    Test that ToolCallingNode with user-provided chat history block merges user and node messages.
    """

    # GIVEN a ToolCallingNode with a user-provided chat history block
    user_chat_history_block = VariablePromptBlock(
        block_type="VARIABLE",
        input_variable="chat_history",
        state=None,
        cache_config=None,
    )

    class TestToolCallingNode(ToolCallingNode):
        ml_model = "gpt-4o-mini"
        blocks = [user_chat_history_block]
        functions = [first_function]
        prompt_inputs = {"chat_history": [ChatMessage(role="USER", text="Hello from user")]}
        max_prompt_iterations = 1

    def generate_prompt_events(*args: Any, **kwargs: Any) -> Iterator[Any]:
        execution_id = str(uuid4())
        events = [
            InitiatedExecutePromptEvent(execution_id=execution_id),
            FulfilledExecutePromptEvent(
                execution_id=execution_id,
                outputs=[StringVellumValue(value="Hello! I can help you.")],
            ),
        ]
        yield from events

    vellum_adhoc_prompt_client.adhoc_execute_prompt_stream.side_effect = generate_prompt_events

    # AND a state
    state = BaseState()

    # WHEN the ToolCallingNode runs
    node = TestToolCallingNode(state=state)
    list(node.run())

    # THEN the API should be called with the correct blocks
    mock_api = vellum_adhoc_prompt_client.adhoc_execute_prompt_stream
    assert mock_api.call_count >= 1

    # AND the blocks should include the user-provided chat_history block
    call_kwargs = mock_api.call_args.kwargs
    blocks = call_kwargs["blocks"]

    chat_history_blocks = [
        block for block in blocks if block.block_type == "VARIABLE" and block.input_variable == "chat_history"
    ]
    assert len(chat_history_blocks) == 1

    # AND the input_values should include the user's chat history
    input_values = call_kwargs["input_values"]
    chat_history_inputs = [
        input_val for input_val in input_values if hasattr(input_val, "key") and input_val.key == "chat_history"
    ]
    assert len(chat_history_inputs) == 1
    assert chat_history_inputs[0].value == [ChatMessage(role="USER", text="Hello from user")]


def test_tool_calling_node_with_generic_type_parameter():
    # GIVEN a custom state class
    class State(BaseState):
        pass

    # AND a ToolCallingNode that uses the generic type parameter
    class TestToolCallingNode(ToolCallingNode[State]):
        ml_model = "gpt-4o-mini"
        blocks = []
        functions = [first_function]
        max_prompt_iterations = 1

    # WHEN we create an instance of the node
    state = State()
    node = TestToolCallingNode(state=state)

    # THEN the node should be created successfully
    assert node is not None
    assert isinstance(node, TestToolCallingNode)
    assert node.state == state


def test_tool_calling_node_workflow_is_dynamic(vellum_adhoc_prompt_client):
    """
    Test workflow_version_exec_config without any mocks to see if that's the issue.
    """

    def generate_prompt_events(*args, **kwargs) -> Iterator[ExecutePromptEvent]:
        execution_id = str(uuid4())

        call_count = vellum_adhoc_prompt_client.adhoc_execute_prompt_stream.call_count
        expected_outputs: List[PromptOutput]
        if call_count == 1:
            expected_outputs = [
                FunctionCallVellumValue(
                    value=FunctionCall(
                        arguments={"var_1": 1, "var_2": 2},
                        id="call_123",
                        name="add_numbers_workflow",
                        state="FULFILLED",
                    ),
                ),
            ]
        else:
            expected_outputs = [StringVellumValue(value="The result is 3")]

        events: List[ExecutePromptEvent] = [
            InitiatedExecutePromptEvent(execution_id=execution_id),
            FulfilledExecutePromptEvent(
                execution_id=execution_id,
                outputs=expected_outputs,
            ),
        ]
        yield from events

    vellum_adhoc_prompt_client.adhoc_execute_prompt_stream.side_effect = generate_prompt_events

    class AddNode(BaseNode):

        class Outputs(BaseNode.Outputs):
            result: int

        def run(self) -> Outputs:
            return self.Outputs(result=1)

    class AddNumbersWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        """
        A simple workflow that adds two numbers.
        """

        graph = AddNode

        class Outputs(BaseWorkflow.Outputs):
            result = AddNode.Outputs.result

    class TestToolCallingNode(ToolCallingNode):
        ml_model = "gpt-4o-mini"
        blocks = []
        functions = [AddNumbersWorkflow]
        prompt_inputs = {}

    # GIVEN a workflow with just a tool calling node
    class ToolCallingWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = TestToolCallingNode

        class Outputs(BaseWorkflow.Outputs):
            text: str = TestToolCallingNode.Outputs.text
            chat_history: List[ChatMessage] = TestToolCallingNode.Outputs.chat_history

    workflow = ToolCallingWorkflow()

    # WHEN the workflow is executed and we capture all events
    events = list(workflow.stream(event_filter=all_workflow_event_filter))

    # AND we should find workflow execution initiated events
    initiated_events = [event for event in events if event.name == "workflow.execution.initiated"]
    assert len(initiated_events) == 3  # Main workflow + tool calling internal + inline workflow

    assert initiated_events[0].body.workflow_definition.is_dynamic is False  # Main workflow
    assert initiated_events[1].body.workflow_definition.is_dynamic is True  # Tool calling internal
    assert initiated_events[2].body.workflow_definition.is_dynamic is True  # Inline workflow


def test_tool_node_preserves_node_exception():
    """Test that tool nodes preserve NodeException error codes and raw_data."""

    def failing_function() -> str:
        raise NodeException(
            message="Custom error",
            code=WorkflowErrorCode.INVALID_INPUTS,
            raw_data={"key": "value"},
        )

    tool_prompt_node = create_tool_prompt_node(
        ml_model="test-model",
        blocks=[],
        functions=[failing_function],
        prompt_inputs=None,
        parameters=DEFAULT_PROMPT_PARAMETERS,
    )

    function_node_class = create_function_node(
        function=failing_function,
        tool_prompt_node=tool_prompt_node,
    )

    state = ToolCallingState(
        meta=StateMeta(
            node_outputs={
                tool_prompt_node.Outputs.results: [
                    FunctionCallVellumValue(
                        value=FunctionCall(
                            arguments={},
                            id="call_123",
                            name="failing_function",
                            state="FULFILLED",
                        ),
                    )
                ],
            },
        )
    )

    function_node = function_node_class(state=state)

    with pytest.raises(NodeException) as exc_info:
        list(function_node.run())

    e = exc_info.value
    assert e.code == WorkflowErrorCode.INVALID_INPUTS
    assert e.raw_data == {"key": "value"}
    assert "Custom error" in e.message


def test_tool_node_error_message_includes_function_name():
    """Test that error messages include the actual function name, not 'wrapper'."""

    # GIVEN a function that raises a regular exception
    def my_tool_function() -> str:
        raise ValueError("Something went wrong")

    # AND a tool prompt node with that function
    tool_prompt_node = create_tool_prompt_node(
        ml_model="test-model",
        blocks=[],
        functions=[my_tool_function],
        prompt_inputs=None,
        parameters=DEFAULT_PROMPT_PARAMETERS,
    )

    function_node_class = create_function_node(
        function=my_tool_function,
        tool_prompt_node=tool_prompt_node,
    )

    # AND a state with a function call
    state = ToolCallingState(
        meta=StateMeta(
            node_outputs={
                tool_prompt_node.Outputs.results: [
                    FunctionCallVellumValue(
                        value=FunctionCall(
                            arguments={},
                            id="call_456",
                            name="my_tool_function",
                            state="FULFILLED",
                        ),
                    )
                ],
            },
        )
    )

    function_node = function_node_class(state=state)

    # WHEN the function node runs and raises an exception
    with pytest.raises(NodeException) as exc_info:
        list(function_node.run())

    # THEN the error message should include the actual function name
    e = exc_info.value
    assert "my_tool_function" in e.message
    assert "wrapper" not in e.message.lower()
    assert "Something went wrong" in e.message


def test_tool_calling_node_400_error_returns_internal_error(vellum_adhoc_prompt_client):
    """
    Test that ToolCallingNode returns INTERNAL_ERROR when the underlying prompt node returns a 400 error.
    """

    # GIVEN a ToolCallingNode with minimal configuration
    class TestToolCallingNode(ToolCallingNode):
        ml_model = "gpt-4o-mini"
        blocks = []
        functions = []
        max_prompt_iterations = 1

    # AND the API client raises a 400 error
    api_error = ApiError(
        status_code=400,
        body={"detail": "Invalid request parameters"},
    )
    vellum_adhoc_prompt_client.adhoc_execute_prompt_stream.side_effect = api_error

    # AND a state
    state = BaseState()

    # WHEN the ToolCallingNode runs
    node = TestToolCallingNode(state=state)

    with pytest.raises(NodeException) as exc_info:
        list(node.run())

    # AND the error code should be INTERNAL_ERROR (not INVALID_INPUTS)
    e = exc_info.value
    assert e.code == WorkflowErrorCode.INTERNAL_ERROR
    assert e.message == "Internal server error"


def test_vellum_integration_node_500_error_feeds_back_to_model(vellum_adhoc_prompt_client, vellum_client):
    """
    Test that NodeException errors from integration tools feed back an error response to the model.
    """
    # GIVEN a VellumIntegrationToolDefinition
    github_create_issue_tool = VellumIntegrationToolDefinition(
        provider=VellumIntegrationProviderType.COMPOSIO,
        integration_name="GITHUB",
        name="create_issue",
        description="Create a new issue in a GitHub repository",
    )

    # AND a ToolCallingNode that uses the integration tool
    class TestToolCallingNode(ToolCallingNode):
        ml_model = "gpt-4o-mini"
        blocks = []
        functions = [github_create_issue_tool]
        max_prompt_iterations = 2

    # AND the integration API raises a 500 error
    vellum_client.integrations.execute_integration_tool.side_effect = ApiError(
        status_code=500,
        body={"detail": "Internal server error occurred while executing the tool."},
    )

    def generate_prompt_events(*args: Any, **kwargs: Any) -> Iterator[Any]:
        execution_id = str(uuid4())
        call_count = vellum_adhoc_prompt_client.adhoc_execute_prompt_stream.call_count

        outputs: List[PromptOutput]
        if call_count == 1:
            outputs = [
                FunctionCallVellumValue(
                    value=FunctionCall(
                        arguments={"owner": "test-owner", "repo": "test-repo", "title": "Test Issue"},
                        id="call_500_error_test",
                        name="create_issue",
                        state="FULFILLED",
                    ),
                )
            ]
        else:
            outputs = [
                StringVellumValue(
                    value="I apologize, but I encountered an error while trying to create the issue. "
                    "The integration service is currently unavailable."
                )
            ]

        events = [
            InitiatedExecutePromptEvent(execution_id=execution_id),
            FulfilledExecutePromptEvent(
                execution_id=execution_id,
                outputs=outputs,
            ),
        ]
        yield from events

    vellum_adhoc_prompt_client.adhoc_execute_prompt_stream.side_effect = generate_prompt_events

    # WHEN the ToolCallingNode runs
    state = ToolCallingState()
    node = TestToolCallingNode(state=state)
    node_outputs = {}
    for output in node.run():
        if output.is_fulfilled:
            node_outputs[output.name] = output.value

    # THEN the node should fulfill successfully
    # AND the chat history should contain a FUNCTION message with the error
    chat_history = cast(List[ChatMessage], node_outputs["chat_history"])
    assert len(chat_history) == 3

    # AND the first message should be the assistant's function call
    assert chat_history[0].role == "ASSISTANT"

    # AND the second message should be the FUNCTION error result
    function_message = chat_history[1]
    assert function_message.role == "FUNCTION"
    assert isinstance(function_message.content, StringChatMessageContent)

    # AND the error payload should contain the error details
    error_data = json.loads(function_message.content.value)
    assert "error" in error_data
    assert error_data["error"]["code"] == "PROVIDER_ERROR"
    assert "Internal server error occurred while executing the tool" in error_data["error"]["message"]

    # AND the third message should be the assistant's response to the error
    assert chat_history[2].role == "ASSISTANT"



from typing import Any


def try_parse_numeric_string(value: Any) -> Any:
    """
    Attempt to parse a string value as a number (int or float).

    This is to support the legacy workflow runner logic where string operands
    should be automatically parsed as numbers when compared with numeric types.

    Returns the parsed number if successful, otherwise returns the original value.
    """
    if not isinstance(value, str):
        return value

    try:
        if "." not in value:
            return int(value)
        return float(value)
    except (ValueError, TypeError):
        return value


def prepare_comparison_operands(lhs: Any, rhs: Any) -> tuple[Any, Any]:
    """
    Prepare operands for comparison by parsing string operands as numbers when appropriate.

    If one operand is a string and the other is a numeric type (int or float),
    attempts to parse the string as a number.

    Returns a tuple of (prepared_lhs, prepared_rhs).
    """
    if isinstance(lhs, str) and isinstance(rhs, (int, float)):
        lhs = try_parse_numeric_string(lhs)
    elif isinstance(rhs, str) and isinstance(lhs, (int, float)):
        rhs = try_parse_numeric_string(rhs)

    return lhs, rhs



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel


class WorkflowResolvedState(UniversalBaseModel):
    """
    The latest execution state of a given Workflow Execution
    """

    trace_id: str
    timestamp: dt.datetime
    span_id: str
    state: typing.Dict[str, typing.Optional[typing.Any]]
    previous_span_id: typing.Optional[str] = None
    previous_trace_id: typing.Optional[str] = None
    root_span_id: typing.Optional[str] = None
    root_trace_id: typing.Optional[str] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from abc import ABC, ABCMeta
from functools import lru_cache
import inspect
import json
import os
from uuid import UUID
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Iterator, Optional, Tuple, Type, cast, get_origin

from vellum.workflows.references.trigger import TriggerAttributeReference
from vellum.workflows.types.utils import get_class_attr_names, infer_types
from vellum.workflows.utils.files import virtual_open
from vellum.workflows.utils.uuids import uuid4_from_hash

if TYPE_CHECKING:
    from vellum.workflows.graph.graph import Graph, GraphTarget
    from vellum.workflows.state.base import BaseState


def _is_annotated(cls: Type, name: str) -> bool:
    annotations = getattr(cls, "__annotations__", {})
    annotation = annotations.get(name)
    if annotation is not None:
        if get_origin(annotation) is ClassVar:
            return False
        return True

    for base in cls.__bases__:
        if _is_annotated(base, name):
            return True

    return False


def _convert_to_relative_module_path(absolute_module_path: str, workflow_root: str) -> str:
    """
    Convert an absolute module path to a relative path from the workflow root.

    Args:
        absolute_module_path: The full module path (e.g., "workflow_id.triggers.scheduled")
        workflow_root: The workflow root module (e.g., "workflow_id")

    Returns:
        Relative module path with leading dot (e.g., ".triggers.scheduled")
    """
    if not absolute_module_path.startswith(workflow_root):
        return ""

    remaining_path = absolute_module_path[len(workflow_root) :]
    if remaining_path.startswith("."):
        return remaining_path
    else:
        return "." + remaining_path


def _get_trigger_id_from_metadata(trigger_class: Type["BaseTrigger"]) -> Optional[UUID]:
    """
    Get the trigger ID from metadata.json for a given trigger class.

    Args:
        trigger_class: The trigger class to look up

    Returns:
        The UUID from metadata.json, or None if not found
    """
    workflow_root = _find_workflow_root_with_metadata(trigger_class.__module__)
    if not workflow_root:
        return None

    trigger_path_to_id_mapping = _get_trigger_path_to_id_mapping(trigger_class.__module__)
    if not trigger_path_to_id_mapping:
        return None

    # Convert module path to relative path and append class name
    # e.g., "root_module.triggers.scheduled" + "ScheduleTrigger" -> ".triggers.scheduled.ScheduleTrigger"
    relative_module_path = _convert_to_relative_module_path(trigger_class.__module__, workflow_root)
    if not relative_module_path:
        return None

    # Append class name to get full trigger path
    relative_trigger_path = f"{relative_module_path}.{trigger_class.__qualname__}"
    return trigger_path_to_id_mapping.get(relative_trigger_path)


def _find_workflow_root_with_metadata(trigger_module: str) -> Optional[str]:
    """
    Find the workflow root module by searching for metadata.json up the module hierarchy.

    Args:
        trigger_module: The trigger's module path (e.g., "workflows.my_workflow.triggers.my_trigger")

    Returns:
        The workflow root module path if found, None otherwise
    """
    module_parts = trigger_module.split(".")

    # Try searching up the module hierarchy for metadata.json
    for i in range(len(module_parts), 0, -1):
        potential_root = ".".join(module_parts[:i])
        module_dir = potential_root.replace(".", os.path.sep)
        metadata_path = os.path.join(module_dir, "metadata.json")

        # Try to open the file using virtual_open to support both regular and virtual filesystems
        # virtual_open checks BaseWorkflowFinder instances before falling back to regular open()
        try:
            file_handle = virtual_open(metadata_path)
            if file_handle is not None:
                file_handle.close()
                return potential_root
        except (FileNotFoundError, OSError):
            pass

    return None


@lru_cache(maxsize=128)
def _get_trigger_path_to_id_mapping(module_path: str) -> Dict[str, UUID]:
    """
    Read trigger path to ID mapping from metadata.json for a given module.

    This function is cached to avoid repeated file reads. It searches up the module
    hierarchy for metadata.json and extracts the trigger_path_to_id_mapping.

    Args:
        module_path: The module path to search from (e.g., "workflows.my_workflow.triggers.my_trigger")

    Returns:
        Dictionary mapping trigger module paths to their UUIDs
    """
    try:
        # Find the workflow root that contains metadata.json
        workflow_root = _find_workflow_root_with_metadata(module_path)
        if not workflow_root:
            return {}

        # Convert module path to file path
        module_dir = workflow_root.replace(".", os.path.sep)
        metadata_path = os.path.join(module_dir, "metadata.json")

        # Use virtual_open to support both regular and virtual environments
        with virtual_open(metadata_path) as f:
            metadata_json = json.load(f)
            trigger_mapping = metadata_json.get("trigger_path_to_id_mapping", {})

            # Convert string IDs to UUIDs
            return {path: UUID(id_str) for path, id_str in trigger_mapping.items()}

    except (FileNotFoundError, json.JSONDecodeError, ValueError, KeyError):
        # If there's any error reading or parsing the file, return empty dict
        return {}


class BaseTriggerMeta(ABCMeta):
    def __new__(mcs, name: str, bases: Tuple[Type, ...], dct: Dict[str, Any]) -> Any:
        if "Display" not in dct:
            for base in reversed(bases):
                if hasattr(base, "Display"):
                    dct["Display"] = type(
                        f"{name}.Display",
                        (base.Display,),
                        {"__module__": dct["__module__"]},
                    )
                    break

        cls = super().__new__(mcs, name, bases, dct)
        trigger_class = cast(Type["BaseTrigger"], cls)

        # Set default ID based on module and class name
        trigger_class.__id__ = uuid4_from_hash(f"{trigger_class.__module__}.{trigger_class.__qualname__}")

        # Try to override with ID from metadata.json if available
        metadata_id = _get_trigger_id_from_metadata(trigger_class)
        if metadata_id:
            trigger_class.__id__ = metadata_id

        return trigger_class

    """
    Metaclass for BaseTrigger that enables class-level >> operator.

    This allows triggers to be used at the class level, similar to nodes:
        ManualTrigger >> MyNode  # Class-level, no instantiation
    """

    def __getattribute__(cls, name: str) -> Any:
        trigger_cls = cast(Type["BaseTrigger"], cls)
        if name.startswith("_"):
            return super().__getattribute__(name)

        try:
            attribute = super().__getattribute__(name)
        except AttributeError as exc:
            if _is_annotated(cls, name):
                attribute = None
            else:
                raise exc

        if inspect.isroutine(attribute) or isinstance(attribute, (property, classmethod, staticmethod)):
            return attribute

        if isinstance(attribute, TriggerAttributeReference):
            return attribute

        if name in cls.__dict__:
            return attribute

        if not _is_annotated(cls, name):
            return attribute

        types = infer_types(cls, name)
        reference = TriggerAttributeReference(name=name, types=types, instance=attribute, trigger_class=trigger_cls)
        return reference

    def __iter__(cls) -> Iterator[TriggerAttributeReference]:
        seen: Dict[str, TriggerAttributeReference] = {}

        for attr_name in get_class_attr_names(cls):
            if attr_name in seen:
                yield seen[attr_name]
                continue

            # Skip ClassVar annotations - they're not trigger attributes
            if not _is_annotated(cls, attr_name):
                continue

            attr_value = getattr(cls, attr_name)
            if isinstance(attr_value, TriggerAttributeReference):
                seen[attr_name] = attr_value
                yield attr_value

    def __rshift__(cls, other: "GraphTarget") -> "Graph":  # type: ignore[misc]
        """
        Enable Trigger class >> Node syntax (class-level only).

        Args:
            other: The target to connect to - can be a Node, Graph, or set of Nodes

        Returns:
            Graph: A graph object with trigger edges

        Examples:
            ManualTrigger >> MyNode
            ManualTrigger >> {NodeA, NodeB}
            ManualTrigger >> (NodeA >> NodeB)
        """
        from vellum.workflows.edges.trigger_edge import TriggerEdge
        from vellum.workflows.graph.graph import Graph
        from vellum.workflows.nodes.bases.base import BaseNode as BaseNodeClass

        # Cast cls to the proper type for TriggerEdge
        trigger_cls = cast(Type["BaseTrigger"], cls)

        if isinstance(other, set):
            # Trigger >> {NodeA, NodeB}
            trigger_edges = []
            graph_items = []
            for item in other:
                if isinstance(item, type) and issubclass(item, BaseNodeClass):
                    trigger_edges.append(TriggerEdge(trigger_cls, item))
                elif isinstance(item, Graph):
                    # Trigger >> {Graph1, Graph2}
                    graph_items.append(item)
                    for entrypoint in item.entrypoints:
                        trigger_edges.append(TriggerEdge(trigger_cls, entrypoint))
                else:
                    raise TypeError(
                        f"Cannot connect trigger to {type(item).__name__}. " f"Expected BaseNode or Graph in set."
                    )

            result_graph = Graph.from_trigger_edges(trigger_edges)

            for graph_item in graph_items:
                result_graph._extend_edges(graph_item.edges)
                result_graph._terminals.update(graph_item._terminals)
                for existing_trigger_edge in graph_item._trigger_edges:
                    if existing_trigger_edge not in result_graph._trigger_edges:
                        result_graph._trigger_edges.append(existing_trigger_edge)

            return result_graph

        elif isinstance(other, Graph):
            # Trigger >> Graph
            edges = [TriggerEdge(trigger_cls, entrypoint) for entrypoint in other.entrypoints]
            result_graph = Graph.from_trigger_edges(edges)
            # Also include the edges from the original graph
            result_graph._extend_edges(other.edges)
            result_graph._terminals = other._terminals
            return result_graph

        elif isinstance(other, type) and issubclass(other, BaseNodeClass):
            # Trigger >> Node
            edge = TriggerEdge(trigger_cls, other)
            return Graph.from_trigger_edge(edge)

        else:
            raise TypeError(
                f"Cannot connect trigger to {type(other).__name__}. " f"Expected BaseNode, Graph, or set of BaseNodes."
            )

    def __rrshift__(cls, other: Any) -> "Graph":
        """
        Prevent Node >> Trigger class syntax.

        Raises:
            TypeError: Always, as this operation is not allowed
        """
        raise TypeError(
            f"Cannot create edge targeting trigger {cls.__name__}. "
            f"Triggers must be at the start of a graph path, not as targets. "
            f"Did you mean: {cls.__name__} >> {other.__name__ if hasattr(other, '__name__') else other}?"
        )


class BaseTrigger(ABC, metaclass=BaseTriggerMeta):
    """
    Base class for workflow triggers - first-class graph elements.

    Triggers define how and when a workflow execution is initiated. They are integrated
    into the workflow graph using the >> operator and can connect to nodes at the class level.

    Examples:
        # Class-level usage (consistent with nodes)
        ManualTrigger >> MyNode
        ManualTrigger >> {NodeA, NodeB}
        ManualTrigger >> (NodeA >> NodeB)

    Subclass Hierarchy:
        - ManualTrigger: Explicit workflow invocation (default)
        - IntegrationTrigger: External service triggers (base for Slack, GitHub, etc.)
        - ScheduledTrigger: Time-based triggers with cron/interval schedules

    Important:
        Triggers can only appear at the start of graph paths. Attempting to create
        edges targeting triggers (Node >> Trigger) will raise a TypeError.

    Note:
        Like nodes, triggers work at the class level only. Do not instantiate triggers.
    """

    __id__: UUID

    class Display:
        """Optional display metadata for visual representation."""

        label: str = "Trigger"
        x: float = 0.0
        y: float = 0.0
        z_index: float = 0.0
        icon: Optional[str] = None
        color: Optional[str] = None

    def __init__(self, **kwargs: Any):
        """
        Initialize trigger with event data.

        Args:
            **kwargs: Arbitrary keyword arguments passed during trigger instantiation.
                     Subclasses may use these to populate trigger attributes.
        """
        self._event_data = kwargs

        for key, value in kwargs.items():
            setattr(self, key, value)

    @classmethod
    def attribute_references(cls) -> Dict[str, "TriggerAttributeReference[Any]"]:
        """Return class-level trigger attribute descriptors keyed by attribute name."""

        return {reference.name: reference for reference in cls}

    def to_trigger_attribute_values(self) -> Dict["TriggerAttributeReference[Any]", Any]:
        """Materialize attribute descriptor/value pairs for this trigger instance."""

        attribute_values: Dict["TriggerAttributeReference[Any]", Any] = {}
        for reference in type(self):
            if hasattr(self, reference.name):
                attribute_values[reference] = getattr(self, reference.name)
            elif type(None) in reference.types:
                attribute_values[reference] = None
            else:
                raise AttributeError(
                    f"{type(self).__name__} did not populate required trigger attribute '{reference.name}'"
                )

        return attribute_values

    def bind_to_state(self, state: "BaseState") -> None:
        """Persist this trigger's attribute values onto the provided state."""

        state.meta.trigger_attributes.update(self.to_trigger_attribute_values())



# This file was auto-generated by Fern from our API Definition.

import typing

from ...core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ...core.request_options import RequestOptions
from ...types.paginated_slim_integration_auth_config_read_list import PaginatedSlimIntegrationAuthConfigReadList
from .raw_client import AsyncRawIntegrationAuthConfigsClient, RawIntegrationAuthConfigsClient


class IntegrationAuthConfigsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawIntegrationAuthConfigsClient(client_wrapper=client_wrapper)

    @property
    def _client_wrapper(self) -> SyncClientWrapper:
        return self._raw_client._client_wrapper

    @property
    def with_raw_response(self) -> RawIntegrationAuthConfigsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawIntegrationAuthConfigsClient
        """
        return self._raw_client

    def list_integration_auth_configs(
        self,
        *,
        expand: typing.Optional[typing.Union[str, typing.Sequence[str]]] = None,
        integration_name: typing.Optional[str] = None,
        integration_provider: typing.Optional[str] = None,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        ordering: typing.Optional[str] = None,
        search: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PaginatedSlimIntegrationAuthConfigReadList:
        """
        List Integration Auth Configs

        Parameters
        ----------
        expand : typing.Optional[typing.Union[str, typing.Sequence[str]]]
            The response fields to expand for more information.
            - 'integration_credentials' expands the list of integration credentials that the authenticated entity has access to for the auth config.

        integration_name : typing.Optional[str]

        integration_provider : typing.Optional[str]

        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        ordering : typing.Optional[str]
            Which field to use when ordering the results.

        search : typing.Optional[str]
            A search term.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PaginatedSlimIntegrationAuthConfigReadList


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.integration_auth_configs.list_integration_auth_configs()
        """
        _response = self._raw_client.list_integration_auth_configs(
            expand=expand,
            integration_name=integration_name,
            integration_provider=integration_provider,
            limit=limit,
            offset=offset,
            ordering=ordering,
            search=search,
            request_options=request_options,
        )
        return _response.data


class AsyncIntegrationAuthConfigsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawIntegrationAuthConfigsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawIntegrationAuthConfigsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawIntegrationAuthConfigsClient
        """
        return self._raw_client

    async def list_integration_auth_configs(
        self,
        *,
        expand: typing.Optional[typing.Union[str, typing.Sequence[str]]] = None,
        integration_name: typing.Optional[str] = None,
        integration_provider: typing.Optional[str] = None,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        ordering: typing.Optional[str] = None,
        search: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PaginatedSlimIntegrationAuthConfigReadList:
        """
        List Integration Auth Configs

        Parameters
        ----------
        expand : typing.Optional[typing.Union[str, typing.Sequence[str]]]
            The response fields to expand for more information.
            - 'integration_credentials' expands the list of integration credentials that the authenticated entity has access to for the auth config.

        integration_name : typing.Optional[str]

        integration_provider : typing.Optional[str]

        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        ordering : typing.Optional[str]
            Which field to use when ordering the results.

        search : typing.Optional[str]
            A search term.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PaginatedSlimIntegrationAuthConfigReadList


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.integration_auth_configs.list_integration_auth_configs()


        asyncio.run(main())
        """
        _response = await self._raw_client.list_integration_auth_configs(
            expand=expand,
            integration_name=integration_name,
            integration_provider=integration_provider,
            limit=limit,
            offset=offset,
            ordering=ordering,
            search=search,
            request_options=request_options,
        )
        return _response.data



# This file was auto-generated by Fern from our API Definition.

from __future__ import annotations

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel, update_forward_refs
from .workflow_event_error import WorkflowEventError
from .workflow_node_result_data import WorkflowNodeResultData


class RejectedWorkflowNodeResultEvent(UniversalBaseModel):
    """
    An event that indicates that the node has rejected its execution.
    """

    id: str
    node_id: str
    node_result_id: str
    state: typing.Literal["REJECTED"] = "REJECTED"
    ts: typing.Optional[dt.datetime] = None
    data: typing.Optional[WorkflowNodeResultData] = None
    source_execution_id: typing.Optional[str] = None
    error: WorkflowEventError

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


from .array_vellum_value import ArrayVellumValue  # noqa: E402, F401, I001

update_forward_refs(RejectedWorkflowNodeResultEvent)



from uuid import UUID
from typing import ClassVar, Generic, Optional, TypeVar

from vellum.workflows.nodes.displayable.final_output_node import FinalOutputNode
from vellum.workflows.types.core import JsonObject
from vellum.workflows.utils.uuids import uuid4_from_hash
from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay
from vellum_ee.workflows.display.nodes.utils import to_kebab_case
from vellum_ee.workflows.display.nodes.vellum.utils import create_node_input
from vellum_ee.workflows.display.types import WorkflowDisplayContext
from vellum_ee.workflows.display.utils.expressions import serialize_value
from vellum_ee.workflows.display.utils.vellum import infer_vellum_variable_type

_FinalOutputNodeType = TypeVar("_FinalOutputNodeType", bound=FinalOutputNode)

NODE_INPUT_KEY = "node_input"


class BaseFinalOutputNodeDisplay(BaseNodeDisplay[_FinalOutputNodeType], Generic[_FinalOutputNodeType]):
    output_name: ClassVar[Optional[str]] = None

    def serialize(self, display_context: WorkflowDisplayContext, **_kwargs) -> JsonObject:
        node = self._node
        node_id = self.node_id

        node_input = create_node_input(
            node_id,
            NODE_INPUT_KEY,
            # Get the pointer that the Terminal Node's output is referencing
            node.Outputs.value.instance,
            display_context,
            self._get_node_input_id(),
        )
        inferred_type = infer_vellum_variable_type(node.Outputs.value)

        return {
            "id": str(node_id),
            "type": "TERMINAL",
            "data": {
                "label": self.label,
                "name": self._get_output_name(),
                "target_handle_id": str(self.get_target_handle_id()),
                "output_id": str(self._get_output_id()),
                "output_type": inferred_type,
                "node_input_id": str(node_input.id),
            },
            "inputs": [node_input.dict()],
            **self.serialize_generic_fields(display_context, exclude=["outputs"]),
            "outputs": [
                {
                    "id": str(self._get_output_id()),
                    "name": node.Outputs.value.name,
                    "type": inferred_type,
                    "value": serialize_value(node_id, display_context, node.Outputs.value.instance),
                }
            ],
        }

    def _get_output_id(self) -> UUID:
        explicit_value = self.output_display.get(self._node.Outputs.value)
        return explicit_value.id if explicit_value else uuid4_from_hash(f"{self.node_id}|output_id")

    def _get_output_name(self) -> str:
        explicit_value = self._get_explicit_node_display_attr("output_name", str)
        return explicit_value if explicit_value else to_kebab_case(self._node.__name__)

    def _get_node_input_id(self) -> UUID:
        explicit_value = self.node_input_ids_by_name.get(NODE_INPUT_KEY)
        return explicit_value if explicit_value else uuid4_from_hash(f"{self.node_id}|node_input_id")



import pytest
import json
import os
import shutil
import subprocess
import tempfile
from unittest.mock import MagicMock
from uuid import uuid4
from typing import Generator

from click.testing import CliRunner
from httpx import Response

from vellum.client.types.docker_service_token import DockerServiceToken
from vellum_cli import main as cli_main


@pytest.fixture
def mock_temp_dir() -> Generator[str, None, None]:
    current_dir = os.getcwd()
    temp_dir = tempfile.mkdtemp()
    os.chdir(temp_dir)

    yield temp_dir

    os.chdir(current_dir)
    shutil.rmtree(temp_dir)


@pytest.fixture
def mock_docker_from_env(mocker):
    return mocker.patch("docker.from_env")


@pytest.fixture
def mock_subprocess_run(mocker):
    return mocker.patch("subprocess.run")


@pytest.mark.usefixtures("vellum_client", "info_log_level")
def test_image_push__self_hosted_happy_path(mock_docker_from_env, mock_subprocess_run, monkeypatch):
    # GIVEN a self hosted vellum api URL env var
    monkeypatch.setenv("VELLUM_API_URL", "mycompany.api.com")
    monkeypatch.setenv("VELLUM_API_KEY", "123456abcdef")

    # Mock Docker client
    mock_docker_client = MagicMock()
    mock_docker_from_env.return_value = mock_docker_client

    mock_subprocess_run.side_effect = [
        subprocess.CompletedProcess(args="", returncode=0, stdout=b"Pruning successful"),
        subprocess.CompletedProcess(
            args="", returncode=0, stdout=b'{"manifests": [{"platform": {"architecture": "amd64"}}]}'
        ),
        subprocess.CompletedProcess(args="", returncode=0, stdout=b"manifest"),
        subprocess.CompletedProcess(args="", returncode=0, stdout=b"sha256:hellosha"),
    ]

    # WHEN the user runs the image push command
    runner = CliRunner()
    result = runner.invoke(cli_main, ["image", "push", "myrepo.net/myimage:latest"])

    # THEN the command exits successfully
    assert result.exit_code == 0, result.output

    # AND gives the success message
    assert "Image successfully pushed" in result.output


@pytest.mark.usefixtures("info_log_level")
def test_image_push__self_hosted_happy_path__workspace_option(
    mock_docker_from_env, mock_subprocess_run, mock_httpx_transport, mock_temp_dir
):
    # GIVEN a workspace config with a new env for url
    with open(os.path.join(mock_temp_dir, "vellum.lock.json"), "w") as f:
        f.write(
            json.dumps(
                {
                    "workspaces": [
                        {
                            "name": "my_workspace",
                            "api_url": "MY_WORKSPACE_VELLUM_API_URL",
                            "api_key": "MY_WORKSPACE_VELLUM_API_KEY",
                        }
                    ]
                }
            )
        )

    # AND a .env file with the workspace api key and url
    with open(os.path.join(mock_temp_dir, ".env"), "w") as f:
        f.write(
            "VELLUM_API_KEY=123456abcdef\n"
            "VELLUM_API_URL=https://api.vellum.ai\n"
            "MY_WORKSPACE_VELLUM_API_KEY=789012ghijkl\n"
            "MY_WORKSPACE_VELLUM_API_URL=https://api.vellum.mycompany.ai\n"
        )

    # AND the Docker client returns the correct response
    mock_docker_client = MagicMock()
    mock_docker_from_env.return_value = mock_docker_client

    mock_subprocess_run.side_effect = [
        subprocess.CompletedProcess(args="", returncode=0, stdout=b"Pruning successful"),
        subprocess.CompletedProcess(
            args="", returncode=0, stdout=b'{"manifests": [{"platform": {"architecture": "amd64"}}]}'
        ),
        subprocess.CompletedProcess(args="", returncode=0, stdout=b"sha256:hellosha"),
    ]

    # AND the vellum client returns the correct response for
    mock_httpx_transport.handle_request.side_effect = [
        # First call to get the docker service token
        Response(
            status_code=200,
            text=json.dumps(
                {
                    "access_token": "345678mnopqr",
                    "organization_id": str(uuid4()),
                    "repository": "myrepo.net",
                }
            ),
        ),
        # Second call to push the image
        Response(
            status_code=200,
            text=json.dumps(
                {
                    "id": str(uuid4()),
                    "name": "myrepo.net/myimage",
                    "visibility": "PRIVATE",
                    "created": "2021-01-01T00:00:00Z",
                    "modified": "2021-01-01T00:00:00Z",
                    "repository": "myrepo.net",
                    "sha": "sha256:hellosha",
                    "tags": [],
                }
            ),
        ),
    ]

    # WHEN the user runs the image push command
    runner = CliRunner()
    result = runner.invoke(cli_main, ["image", "push", "myrepo.net/myimage:latest", "--workspace", "my_workspace"])

    # THEN the command exits successfully
    assert result.exit_code == 0, (result.output, str(result.exception))

    # AND gives the success message
    assert "Image successfully pushed" in result.output

    # AND the vellum client was called with the correct api key and url
    request = mock_httpx_transport.handle_request.call_args[0][0]
    assert request.headers["X-API-KEY"] == "789012ghijkl", result.stdout
    assert str(request.url) == "https://api.vellum.mycompany.ai/v1/container-images/push"


@pytest.mark.usefixtures("vellum_client", "mock_subprocess_run")
def test_image_push__self_hosted_blocks_repo(mock_docker_from_env, monkeypatch):
    # GIVEN a self hosted vellum api URL env var
    monkeypatch.setenv("VELLUM_API_URL", "mycompany.api.com")

    # Mock Docker client
    mock_docker_client = MagicMock()
    mock_docker_from_env.return_value = mock_docker_client

    # WHEN the user runs the image push command
    runner = CliRunner()
    result = runner.invoke(cli_main, ["image", "push", "myimage"])

    # THEN the command exits unsuccessfully
    assert result.exit_code == 1, result.output

    # AND gives the error message for self hosted installs not including the repo
    assert "For adding images to your self hosted install you must include" in result.output


@pytest.mark.usefixtures("info_log_level")
def test_image_push_with_source_success(
    mock_docker_from_env, mock_subprocess_run, vellum_client, monkeypatch, mock_temp_dir
):
    monkeypatch.setenv("VELLUM_API_URL", "https://api.vellum.ai")
    monkeypatch.setenv("VELLUM_API_KEY", "123456abcdef")

    dockerfile_path = os.path.join(mock_temp_dir, "Dockerfile")
    with open(dockerfile_path, "w") as f:
        f.write("FROM alpine:latest\n")

    mock_docker_client = MagicMock()
    mock_docker_from_env.return_value = mock_docker_client
    mock_docker_client.images.push.return_value = [b'{"status": "Pushed"}']

    mock_subprocess_run.side_effect = [
        subprocess.CompletedProcess(args="", returncode=0, stdout=b"Pruning successful"),
        subprocess.CompletedProcess(args="", returncode=0, stdout=b"Build successful"),
        subprocess.CompletedProcess(
            args="", returncode=0, stdout=b'{"manifests": [{"platform": {"architecture": "amd64"}}]}'
        ),
        subprocess.CompletedProcess(args="", returncode=0, stdout=b"sha256:hellosha"),
    ]

    vellum_client.container_images.docker_service_token.return_value = DockerServiceToken(
        access_token="345678mnopqr", organization_id="test-org", repository="myrepo.net"
    )

    runner = CliRunner()
    result = runner.invoke(cli_main, ["image", "push", "myimage:latest", "--source", dockerfile_path])

    assert result.exit_code == 0, result.output

    build_call = mock_subprocess_run.call_args_list[1]
    assert build_call[0][0] == [
        "docker",
        "buildx",
        "build",
        "-f",
        dockerfile_path,
        "--platform=linux/amd64",
        "-t",
        "myimage:latest",
        ".",
    ]

    assert "Docker build completed successfully" in result.output
    assert "Image successfully pushed" in result.output


@pytest.mark.usefixtures("mock_docker_from_env", "mock_subprocess_run", "vellum_client")
def test_image_push_with_source_dockerfile_not_exists(monkeypatch, mock_temp_dir):
    monkeypatch.setenv("VELLUM_API_URL", "https://api.vellum.ai")
    monkeypatch.setenv("VELLUM_API_KEY", "123456abcdef")

    nonexistent_dockerfile = os.path.join(mock_temp_dir, "nonexistent_dockerfile")

    runner = CliRunner()
    result = runner.invoke(cli_main, ["image", "push", "myimage:latest", "--source", nonexistent_dockerfile])

    assert result.exit_code == 1
    assert "Dockerfile does not exist" in result.output


@pytest.mark.usefixtures("mock_docker_from_env", "vellum_client")
def test_image_push_with_source_build_fails(mock_subprocess_run, monkeypatch, mock_temp_dir):
    monkeypatch.setenv("VELLUM_API_URL", "https://api.vellum.ai")
    monkeypatch.setenv("VELLUM_API_KEY", "123456abcdef")

    dockerfile_path = os.path.join(mock_temp_dir, "Dockerfile")
    with open(dockerfile_path, "w") as f:
        f.write("FROM alpine:latest\n")

    mock_subprocess_run.side_effect = [
        subprocess.CompletedProcess(args="", returncode=0, stdout=b"Pruning successful"),
        subprocess.CompletedProcess(args="", returncode=1, stderr=b"Build failed: missing dependency"),
    ]

    runner = CliRunner()
    result = runner.invoke(cli_main, ["image", "push", "myimage:latest", "--source", dockerfile_path])

    assert result.exit_code == 1
    assert "Docker build failed" in result.output
    assert "Build failed: missing dependency" in result.output


@pytest.mark.usefixtures("vellum_client", "info_log_level")
def test_image_push_includes_docker_prune(mock_docker_from_env, mock_subprocess_run, monkeypatch):
    """
    Tests that image_push_command calls docker image prune before validation.
    """
    # GIVEN a self hosted vellum api URL env var
    monkeypatch.setenv("VELLUM_API_URL", "mycompany.api.com")
    monkeypatch.setenv("VELLUM_API_KEY", "123456abcdef")

    # AND a mock Docker client
    mock_docker_client = MagicMock()
    mock_docker_from_env.return_value = mock_docker_client

    mock_subprocess_run.side_effect = [
        subprocess.CompletedProcess(args="", returncode=0, stdout=b"Pruning successful"),
        subprocess.CompletedProcess(
            args="", returncode=0, stdout=b'{"manifests": [{"platform": {"architecture": "amd64"}}]}'
        ),
        subprocess.CompletedProcess(args="", returncode=0, stdout=b"manifest"),
        subprocess.CompletedProcess(args="", returncode=0, stdout=b"sha256:hellosha"),
    ]

    # WHEN the user runs the image push command with any image
    runner = CliRunner()
    result = runner.invoke(cli_main, ["image", "push", "myrepo.net/myimage:latest"])

    # THEN the command exits successfully
    assert result.exit_code == 0, result.output

    prune_call = mock_subprocess_run.call_args_list[0]
    assert prune_call[0][0] == [
        "docker",
        "image",
        "prune",
        "--all",
        "-f",
        "--filter",
        "label=image-type=python-workflow-runtime",
    ]

    # AND the success message includes pruning completion
    assert "Docker image pruning completed successfully" in result.output


@pytest.mark.usefixtures("vellum_client", "info_log_level")
def test_image_push_continues_if_prune_fails(mock_docker_from_env, mock_subprocess_run, monkeypatch):
    """
    Tests that image_push_command continues if docker image prune fails.
    """
    # GIVEN a self hosted vellum api URL env var
    monkeypatch.setenv("VELLUM_API_URL", "mycompany.api.com")
    monkeypatch.setenv("VELLUM_API_KEY", "123456abcdef")

    # AND a mock Docker client
    mock_docker_client = MagicMock()
    mock_docker_from_env.return_value = mock_docker_client

    mock_subprocess_run.side_effect = [
        subprocess.CalledProcessError(
            1, ["docker", "image", "prune", "--all", "-f", "--filter", "label=image-type=python-workflow-runtime"]
        ),
        subprocess.CompletedProcess(
            args="", returncode=0, stdout=b'{"manifests": [{"platform": {"architecture": "amd64"}}]}'
        ),
        subprocess.CompletedProcess(args="", returncode=0, stdout=b"manifest"),
        subprocess.CompletedProcess(args="", returncode=0, stdout=b"sha256:hellosha"),
    ]

    # WHEN the user runs the image push command with any image
    runner = CliRunner()
    result = runner.invoke(cli_main, ["image", "push", "myrepo.net/myimage:latest"])

    # THEN the command exits successfully despite pruning failure
    assert result.exit_code == 0, result.output

    # AND the warning message about pruning failure is shown
    assert "Docker image pruning failed" in result.output

    # AND the success message is still shown
    assert "Image successfully pushed" in result.output



import io
import zipfile


def zip_file_map(file_map: dict[str, str]) -> bytes:
    """
    Create a zip file from a dictionary of file names to content.

    Args:
        file_map: Dictionary mapping file names to their content

    Returns:
        Bytes representing the zip file
    """
    zip_buffer = io.BytesIO()

    with zipfile.ZipFile(zip_buffer, "w", zipfile.ZIP_DEFLATED) as zip_file:
        for filename, content in file_map.items():
            zip_file.writestr(filename, content)

    zip_bytes = zip_buffer.getvalue()
    zip_buffer.close()

    return zip_bytes


def extract_zip_files(zip_bytes: bytes) -> dict[str, str]:
    """
    Extract files from a zip archive.

    Args:
        zip_bytes: Bytes representing the zip file

    Returns:
        Dictionary mapping file names to their content
    """
    zip_buffer = io.BytesIO(zip_bytes)
    extracted_files = {}

    with zipfile.ZipFile(zip_buffer) as zip_file:
        for file_name in zip_file.namelist():
            with zip_file.open(file_name) as source:
                content = source.read().decode("utf-8")
                extracted_files[file_name] = content

    return extracted_files



from decimal import Decimal
from uuid import UUID
from typing import ClassVar, Generic, List, Optional, Union

from vellum import (
    NotFoundError,
    SearchFiltersRequest,
    SearchRequestOptionsRequest,
    SearchResponse,
    SearchResult,
    SearchResultMergingRequest,
    SearchWeightsRequest,
)
from vellum.client.core import ApiError, RequestOptions
from vellum.workflows.errors import WorkflowErrorCode
from vellum.workflows.exceptions import NodeException
from vellum.workflows.nodes.bases import BaseNode
from vellum.workflows.nodes.displayable.bases.types import SearchFilters
from vellum.workflows.outputs import BaseOutputs
from vellum.workflows.types import MergeBehavior
from vellum.workflows.types.generics import StateType

DEFAULT_SEARCH_WEIGHTS = 0.8
DEFAULT_SEARCH_LIMIT = 8


def get_default_results_merging() -> SearchResultMergingRequest:
    return SearchResultMergingRequest(enabled=True)


class BaseSearchNode(BaseNode[StateType], Generic[StateType]):
    """
    Used to perform a hybrid search against a Document Index in Vellum.

    document_index: Union[UUID, str] - Either the UUID or name of the Vellum Document Index that you'd like to search
        against
    query: str - The query to search for
    limit: Optional[int] = None - The maximum number of results to return.
    weights: Optional[SearchWeightsRequest] = None - The weights to use for the search. Must add up to 1.0.
    result_merging: Optional[SearchResultMergingRequest] = None - The configuration for merging results.
    filters: Optional[SearchFiltersRequest] = None - The filters to apply to the search.
    options: Optional[SearchRequestOptionsRequest] = None - [DEPRECATED] Runtime configuration for the search
    request_options: Optional[RequestOptions] = None - The request options to use for the search
    """

    # The query to search for.
    query: ClassVar[str]

    # The Document Index to Search against. Identified by either its UUID or its name.
    document_index: ClassVar[Union[UUID, str]]

    # The maximum number of results to return.
    limit: ClassVar[Optional[int]] = None

    # The weights to use for the search. Must add up to 1.0.
    weights: ClassVar[Optional[SearchWeightsRequest]] = None

    # The configuration for merging results.
    result_merging: ClassVar[Optional[SearchResultMergingRequest]] = None

    # The filters to apply to the search.
    filters: ClassVar[Optional[SearchFilters]] = None

    # Ideally we could reuse node descriptors to derive other node descriptor values. Two action items are
    # blocking us from doing so in this use case:
    # 1. Node Descriptor resolution during runtime - https://app.shortcut.com/vellum/story/4781
    # 2. Math operations between descriptors - https://app.shortcut.com/vellum/story/4782
    # search_weights = DEFAULT_SEARCH_WEIGHTS
    # Deprecated: Use the top level `limit`, `weights`, `result_merging`, and `filters` attributes instead
    options = SearchRequestOptionsRequest(
        limit=DEFAULT_SEARCH_LIMIT,
        weights=SearchWeightsRequest(
            semantic_similarity=DEFAULT_SEARCH_WEIGHTS,
            keywords=float(Decimal("1.0") - Decimal(str(DEFAULT_SEARCH_WEIGHTS))),
        ),
        result_merging=get_default_results_merging(),
        filters=SearchFiltersRequest(
            external_ids=None,
            metadata=None,
        ),
    )

    request_options: Optional[RequestOptions] = None

    class Trigger(BaseNode.Trigger):
        merge_behavior = MergeBehavior.AWAIT_ANY

    class Outputs(BaseOutputs):
        """
        The outputs of the SearchNode.

        results: List[SearchResult] - The raw search results
        """

        results: List[SearchResult]

    def _perform_search(self) -> SearchResponse:
        try:
            return self._context.vellum_client.search(
                query=self.query,
                document_index=str(self.document_index),
                options=self._get_options_request(),
            )
        except NotFoundError:
            raise NodeException(
                message=f"Document Index '{self.document_index}' not found",
                code=WorkflowErrorCode.INVALID_INPUTS,
            )
        except ApiError as e:
            raw_data = e.body if isinstance(e.body, dict) else None
            if e.status_code and e.status_code == 403 and isinstance(e.body, dict):
                raise NodeException(
                    message=e.body.get("detail", "Provider credentials is missing or unavailable"),
                    code=WorkflowErrorCode.PROVIDER_CREDENTIALS_UNAVAILABLE,
                    raw_data=raw_data,
                )
            elif e.status_code and e.status_code >= 400 and e.status_code < 500 and isinstance(e.body, dict):
                raise NodeException(
                    message=e.body.get(
                        "detail", f"An error occurred while searching against Document Index '{self.document_index}'"
                    ),
                    code=WorkflowErrorCode.INVALID_INPUTS,
                    raw_data=raw_data,
                ) from e
            raise NodeException(
                message=f"An error occurred while searching against Document Index '{self.document_index}'",
                code=WorkflowErrorCode.INTERNAL_ERROR,
                raw_data=raw_data,
            ) from e

    def _get_options_request(self) -> SearchRequestOptionsRequest:
        return SearchRequestOptionsRequest(
            limit=self.limit if self.limit is not None else self.options.limit,
            weights=self.weights if self.weights is not None else self.options.weights,
            result_merging=self.result_merging if self.result_merging is not None else self.options.result_merging,
            filters=self._get_filters_request(),
        )

    def _get_filters_request(self) -> Optional[SearchFiltersRequest]:
        if self.filters is None:
            return self.options.filters

        return SearchFiltersRequest(
            external_ids=self.filters.external_ids,
            metadata=self.filters.metadata.to_request() if self.filters.metadata is not None else None,
        )

    def run(self) -> Outputs:
        response = self._perform_search()
        return self.Outputs(results=response.results)



import uuid
from unittest.mock import Mock
from vellum.client.utils import convert_input_variable_to_uuid


def test_convert_input_variable_to_uuid_with_valid_uuid():
    """Test convert_input_variable_to_uuid with valid UUID passes through unchanged."""
    uuid_value = "123e4567-e89b-12d3-a456-426614174000"

    # GIVEN a valid UUID
    # WHEN the function is called
    result = convert_input_variable_to_uuid(uuid_value)

    # THEN the result should be the original UUID unchanged
    assert result == uuid_value


def test_convert_input_variable_to_uuid_with_invalid_uuid_converts():
    """Test convert_input_variable_to_uuid with invalid UUID gets converted to UUID."""
    non_uuid_value = "some_variable_name"
    executable_id = "test_executable_123"

    # GIVEN an invalid UUID with context
    mock_info = Mock()
    mock_info.context = {"executable_id": executable_id}

    # WHEN the function is called
    result = convert_input_variable_to_uuid(non_uuid_value, mock_info)

    # THEN it should return a different value
    assert result != non_uuid_value

    # AND it's a valid UUID
    uuid.UUID(result)  # This will raise ValueError if not a valid UUID



from vellum.workflows.inputs import BaseInputs
from vellum.workflows.nodes import BaseNode, InlineSubworkflowNode
from vellum.workflows.outputs.base import BaseOutputs
from vellum.workflows.state.base import BaseState
from vellum.workflows.workflows.base import BaseWorkflow
from vellum.workflows.workflows.event_filters import workflow_sandbox_event_filter


class NestedInputs(BaseInputs):
    value: str


class NestedNode(BaseNode):
    value = NestedInputs.value

    class Outputs(BaseOutputs):
        result: str

    def run(self) -> Outputs:
        return self.Outputs(result=f"nested: {self.value}")


class NestedWorkflow(BaseWorkflow[NestedInputs, BaseState]):
    graph = NestedNode

    class Outputs(BaseOutputs):
        result = NestedNode.Outputs.result


class ParentInputs(BaseInputs):
    value: str


class SubworkflowNode(InlineSubworkflowNode):
    subworkflow_inputs = {
        "value": ParentInputs.value,
    }
    subworkflow = NestedWorkflow


class ParentWorkflow(BaseWorkflow[ParentInputs, BaseState]):
    graph = SubworkflowNode

    class Outputs(BaseOutputs):
        result = SubworkflowNode.Outputs.result


def test_workflow_sandbox_event_filter__filters_nested_workflow_snapshotted_events():
    """
    Tests that workflow_sandbox_event_filter filters out snapshotted events from nested workflows.
    """

    workflow = ParentWorkflow()

    # WHEN we stream the workflow with workflow_sandbox_event_filter
    events = list(
        workflow.stream(
            inputs=ParentInputs(value="test"),
            event_filter=workflow_sandbox_event_filter,
        )
    )

    snapshotted_events = [e for e in events if e.name == "workflow.execution.snapshotted"]
    assert len(snapshotted_events) > 0

    for event in snapshotted_events:
        assert event.workflow_definition == ParentWorkflow


def test_workflow_sandbox_event_filter__includes_root_workflow_snapshotted_events():
    """
    Tests that workflow_sandbox_event_filter includes snapshotted events from the root workflow.
    """

    class SimpleNode(BaseNode):
        class Outputs(BaseOutputs):
            result: str = "simple"

        def run(self) -> Outputs:
            return self.Outputs()

    class SimpleWorkflow(BaseWorkflow[BaseInputs, BaseState]):
        graph = SimpleNode

        class Outputs(BaseOutputs):
            result = SimpleNode.Outputs.result

    workflow = SimpleWorkflow()

    # WHEN we stream the workflow with workflow_sandbox_event_filter
    events = list(
        workflow.stream(
            inputs=BaseInputs(),
            event_filter=workflow_sandbox_event_filter,
        )
    )

    snapshotted_events = [e for e in events if e.name == "workflow.execution.snapshotted"]
    assert len(snapshotted_events) > 0

    for event in snapshotted_events:
        assert event.workflow_definition == SimpleWorkflow


def test_workflow_sandbox_event_filter__includes_nested_workflow_non_snapshotted_events():
    """
    Tests that workflow_sandbox_event_filter includes non-snapshotted events from nested workflows.
    """

    workflow = ParentWorkflow()

    # WHEN we stream the workflow with workflow_sandbox_event_filter
    events = list(
        workflow.stream(
            inputs=ParentInputs(value="test"),
            event_filter=workflow_sandbox_event_filter,
        )
    )

    nested_workflow_events = [
        e for e in events if hasattr(e, "workflow_definition") and e.workflow_definition == NestedWorkflow
    ]
    assert len(nested_workflow_events) > 0

    for event in nested_workflow_events:
        assert event.name != "workflow.execution.snapshotted"



import inspect
from typing import Any, Generic, TypeVar

from vellum.workflows.nodes.core.retry_node.node import RetryNode
from vellum.workflows.nodes.utils import ADORNMENT_MODULE_NAME
from vellum.workflows.references.output import OutputReference
from vellum.workflows.types.core import JsonArray, JsonObject
from vellum.workflows.utils.uuids import uuid4_from_hash
from vellum.workflows.workflows.base import BaseWorkflow
from vellum_ee.workflows.display.nodes.get_node_display_class import get_node_display_class
from vellum_ee.workflows.display.nodes.types import NodeOutputDisplay
from vellum_ee.workflows.display.nodes.vellum.base_adornment_node import BaseAdornmentNodeDisplay
from vellum_ee.workflows.display.types import WorkflowDisplayContext
from vellum_ee.workflows.display.utils.expressions import serialize_value

_RetryNodeType = TypeVar("_RetryNodeType", bound=RetryNode)


class BaseRetryNodeDisplay(BaseAdornmentNodeDisplay[_RetryNodeType], Generic[_RetryNodeType]):
    def serialize(self, display_context: WorkflowDisplayContext, **_kwargs: Any) -> JsonObject:
        node = self._node
        node_id = self.node_id

        attributes: JsonArray = []
        for attribute in node:
            if inspect.isclass(attribute.instance) and issubclass(attribute.instance, BaseWorkflow):
                # We don't need to serialize attributes that are workflows
                continue

            id = str(uuid4_from_hash(f"{node_id}|{attribute.name}"))
            attributes.append(
                {
                    "id": id,
                    "name": attribute.name,
                    "value": serialize_value(node_id, display_context, attribute.instance),
                }
            )

        adornment: JsonObject = {
            "id": str(node_id),
            "label": self.label,
            "base": self.get_base().dict(),
            "attributes": attributes,
        }

        serialized_node = super().serialize(
            display_context,
            adornment=adornment,
        )

        if serialized_node["type"] == "GENERIC":
            return serialized_node

        serialized_node_definition = serialized_node.get("definition")
        if isinstance(serialized_node_definition, dict):
            serialized_node_definition_module = serialized_node_definition.get("module")
            if isinstance(serialized_node_definition_module, list):
                serialized_node_definition_module.extend(
                    [
                        serialized_node_definition["name"],
                        ADORNMENT_MODULE_NAME,
                    ]
                )
                serialized_node_definition["name"] = node.__name__

        return serialized_node

    def get_node_output_display(self, output: OutputReference) -> NodeOutputDisplay:
        inner_node = self._node.__wrapped_node__
        if not inner_node:
            return super().get_node_output_display(output)

        node_display_class = get_node_display_class(inner_node)
        node_display = node_display_class()

        inner_output = getattr(inner_node.Outputs, output.name)
        return node_display.get_node_output_display(inner_output)



from typing import TYPE_CHECKING

from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.types.core import VellumSecret

if TYPE_CHECKING:
    from vellum.workflows.state.base import BaseState


class VellumSecretReference(BaseDescriptor[VellumSecret]):
    def __init__(self, name: str):
        super().__init__(name=name, types=(VellumSecret,))

    def resolve(self, state: "BaseState") -> VellumSecret:
        return VellumSecret(name=self.name)
